# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'ui_options_tab.ui'
#
# Created by: PyQt5 UI code generator 5.12.2
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_options(object):
    def setupUi(self, options):
        options.setObjectName("options")
        options.resize(761, 737)
        self.label_2 = QtWidgets.QLabel(options)
        self.label_2.setGeometry(QtCore.QRect(10, 10, 118, 29))
        self.label_2.setObjectName("label_2")
        self.line = QtWidgets.QFrame(options)
        self.line.setGeometry(QtCore.QRect(0, 50, 741, 21))
        self.line.setFrameShape(QtWidgets.QFrame.HLine)
        self.line.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line.setObjectName("line")
        self.label = QtWidgets.QLabel(options)
        self.label.setGeometry(QtCore.QRect(20, 20, 16, 29))
        self.label.setText("")
        self.label.setObjectName("label")
        self.widgetFlipButtons = QtWidgets.QWidget(options)
        self.widgetFlipButtons.setGeometry(QtCore.QRect(10, 80, 344, 161))
        self.widgetFlipButtons.setObjectName("widgetFlipButtons")
        self.label_3 = QtWidgets.QLabel(self.widgetFlipButtons)
        self.label_3.setGeometry(QtCore.QRect(9, 9, 85, 17))
        self.label_3.setObjectName("label_3")
        self.layoutWidget = QtWidgets.QWidget(self.widgetFlipButtons)
        self.layoutWidget.setGeometry(QtCore.QRect(10, 30, 306, 100))
        self.layoutWidget.setObjectName("layoutWidget")
        self.gridLayout = QtWidgets.QGridLayout(self.layoutWidget)
        self.gridLayout.setContentsMargins(0, 0, 0, 0)
        self.gridLayout.setObjectName("gridLayout")
        self.label_4 = QtWidgets.QLabel(self.layoutWidget)
        self.label_4.setObjectName("label_4")
        self.gridLayout.addWidget(self.label_4, 0, 1, 1, 1)
        self.label_5 = QtWidgets.QLabel(self.layoutWidget)
        self.label_5.setObjectName("label_5")
        self.gridLayout.addWidget(self.label_5, 0, 2, 1, 1)
        self.label_6 = QtWidgets.QLabel(self.layoutWidget)
        self.label_6.setObjectName("label_6")
        self.gridLayout.addWidget(self.label_6, 0, 3, 1, 1)
        self.label_7 = QtWidgets.QLabel(self.layoutWidget)
        self.label_7.setObjectName("label_7")
        self.gridLayout.addWidget(self.label_7, 1, 0, 1, 1)
        self.checkBoxSagittalFlipX = QtWidgets.QCheckBox(self.layoutWidget)
        self.checkBoxSagittalFlipX.setText("")
        self.checkBoxSagittalFlipX.setObjectName("checkBoxSagittalFlipX")
        self.gridLayout.addWidget(self.checkBoxSagittalFlipX, 1, 1, 1, 1)
        self.checkBoxCoronalFlipX = QtWidgets.QCheckBox(self.layoutWidget)
        self.checkBoxCoronalFlipX.setText("")
        self.checkBoxCoronalFlipX.setObjectName("checkBoxCoronalFlipX")
        self.gridLayout.addWidget(self.checkBoxCoronalFlipX, 1, 2, 1, 1)
        self.checkBoxAxialFlipx = QtWidgets.QCheckBox(self.layoutWidget)
        self.checkBoxAxialFlipx.setText("")
        self.checkBoxAxialFlipx.setObjectName("checkBoxAxialFlipx")
        self.gridLayout.addWidget(self.checkBoxAxialFlipx, 1, 3, 1, 1)
        self.label_10 = QtWidgets.QLabel(self.layoutWidget)
        self.label_10.setObjectName("label_10")
        self.gridLayout.addWidget(self.label_10, 2, 0, 1, 1)
        self.checkBoxSagittalFlipY = QtWidgets.QCheckBox(self.layoutWidget)
        self.checkBoxSagittalFlipY.setText("")
        self.checkBoxSagittalFlipY.setObjectName("checkBoxSagittalFlipY")
        self.gridLayout.addWidget(self.checkBoxSagittalFlipY, 2, 1, 1, 1)
        self.checkBoxCoronalFlipY = QtWidgets.QCheckBox(self.layoutWidget)
        self.checkBoxCoronalFlipY.setText("")
        self.checkBoxCoronalFlipY.setObjectName("checkBoxCoronalFlipY")
        self.gridLayout.addWidget(self.checkBoxCoronalFlipY, 2, 2, 1, 1)
        self.checkBoxAxialFlipY = QtWidgets.QCheckBox(self.layoutWidget)
        self.checkBoxAxialFlipY.setText("")
        self.checkBoxAxialFlipY.setObjectName("checkBoxAxialFlipY")
        self.gridLayout.addWidget(self.checkBoxAxialFlipY, 2, 3, 1, 1)
        self.label_8 = QtWidgets.QLabel(self.layoutWidget)
        self.label_8.setObjectName("label_8")
        self.gridLayout.addWidget(self.label_8, 3, 0, 1, 1)
        self.checkBoxSagittalFlipZ = QtWidgets.QCheckBox(self.layoutWidget)
        self.checkBoxSagittalFlipZ.setText("")
        self.checkBoxSagittalFlipZ.setObjectName("checkBoxSagittalFlipZ")
        self.gridLayout.addWidget(self.checkBoxSagittalFlipZ, 3, 1, 1, 1)
        self.checkBoxCoronalFlipZ = QtWidgets.QCheckBox(self.layoutWidget)
        self.checkBoxCoronalFlipZ.setText("")
        self.checkBoxCoronalFlipZ.setObjectName("checkBoxCoronalFlipZ")
        self.gridLayout.addWidget(self.checkBoxCoronalFlipZ, 3, 2, 1, 1)
        self.checkBoxAxialFlipZ = QtWidgets.QCheckBox(self.layoutWidget)
        self.checkBoxAxialFlipZ.setText("")
        self.checkBoxAxialFlipZ.setObjectName("checkBoxAxialFlipZ")
        self.gridLayout.addWidget(self.checkBoxAxialFlipZ, 3, 3, 1, 1)
        self.checkBoxImpcView = QtWidgets.QCheckBox(options)
        self.checkBoxImpcView.setGeometry(QtCore.QRect(20, 270, 62, 20))
        self.checkBoxImpcView.setText("")
        self.checkBoxImpcView.setObjectName("checkBoxImpcView")
        self.textBrowser = QtWidgets.QTextBrowser(options)
        self.textBrowser.setGeometry(QtCore.QRect(50, 270, 256, 151))
        self.textBrowser.setObjectName("textBrowser")
        self.lineEditShowLabel = QtWidgets.QLineEdit(options)
        self.lineEditShowLabel.setGeometry(QtCore.QRect(30, 490, 113, 25))
        self.lineEditShowLabel.setObjectName("lineEditShowLabel")
        self.label_9 = QtWidgets.QLabel(options)
        self.label_9.setGeometry(QtCore.QRect(10, 460, 461, 17))
        self.label_9.setObjectName("label_9")

        self.retranslateUi(options)
        QtCore.QMetaObject.connectSlotsByName(options)

    def retranslateUi(self, options):
        _translate = QtCore.QCoreApplication.translate
        options.setWindowTitle(_translate("options", "Form"))
        self.label_2.setText(_translate("options", "Options"))
        self.label_3.setText(_translate("options", "Orientations"))
        self.label_4.setText(_translate("options", "Sagittal"))
        self.label_5.setText(_translate("options", "coronal"))
        self.label_6.setText(_translate("options", "axial"))
        self.label_7.setText(_translate("options", "Flip horizontal"))
        self.label_10.setText(_translate("options", "flip vertical"))
        self.label_8.setText(_translate("options", "Reverse slice ordering"))
        self.textBrowser.setHtml(_translate("options", "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//EN\" \"http://www.w3.org/TR/REC-html40/strict.dtd\">\n"
"<html><head><meta name=\"qrichtext\" content=\"1\" /><style type=\"text/css\">\n"
"p, li { white-space: pre-wrap; }\n"
"</style></head><body style=\" font-family:\'Ubuntu\'; font-size:11pt; font-weight:400; font-style:normal;\">\n"
"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><span style=\" font-weight:600;\">IMPC view. </span></p>\n"
"<p style=\"-qt-paragraph-type:empty; margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\"><br /></p>\n"
"<p style=\" margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\">Checking this button with ensure that data downloaded from the DCC will be in the agreed orientation. For other data you may have to use the individual flip options.</p></body></html>"))
        self.label_9.setText(_translate("options", "Show only these lables (space-seperated list of label numbers)"))





from PyQt5 import QtGui
import pyqtgraph as pg
import numpy as np
from .layer import Layer


class HeatmapLayer(Layer):
    def __init__(self, *args):
        super(HeatmapLayer, self).__init__(*args)
        self.neg_image_item = pg.ImageItem(autoLevels=False)
        self.neg_image_item.setCompositionMode(QtGui.QPainter.CompositionMode_SourceOver)
        self.neg_image_item.setLookupTable(self.lt._hot_red_blue()[1])
        self.image_items.append(self.neg_image_item)

        self.pos_image_item = pg.ImageItem(autoLevels=False)
        self.pos_image_item.setCompositionMode(QtGui.QPainter.CompositionMode_SourceOver)
        self.pos_image_item.setLookupTable(self.lt._hot_red_blue()[0])
        self.image_items.append(self.pos_image_item)

    def update(self, auto_levels=False):
        """
        :param orientation:
        :param auto_levels:
        :return:
        """
        if self.vol and self.vol != 'None':
            self.neg_image_item.setLookupTable(self.vol.negative_lut)
            self.pos_image_item.setLookupTable(self.vol.positive_lut)
            self.neg_image_item.setLevels(self.vol.neg_levels, update=False)
            self.pos_image_item.setLevels(self.vol.pos_levels, update=False)
            self.reload()

    def reload(self):
        """
        """
        # need to work out if this is needed. It currently puts heatmap data in wrong flip at first view
        if self.vol:
            self.set_slice(self.parent.current_slice_idx)
           
    def set_volume(self, volname):
        """
        :param vol, Volume object from model.py
        """
        if volname == "None":
            self.volume_label_signal.emit("None")
            self.clear()
            self.vol = None
            return

        self.volume_label_signal.emit(volname)
        self.vol = self.model.getvol(volname)
        self.neg_image_item.setLookupTable(self.vol.negative_lut)
        self.pos_image_item.setLookupTable(self.vol.positive_lut)
        self.set_slice(self.parent.current_slice_idx)

    def set_slice(self, index: int):
        """Set the heatmap layer to the specified index

        Parameters
        ----------
        index: int
            the index to set

        """
        opacity = 1.0 if self.isvisible else 0.0
        if self.vol and self.vol != "None":

            flip_x, flip_y, flip_z = self.parent.get_flips()

            try:
                slices = self.vol.get_data(self.parent.orientation, index - 1,
                                           flip_x, flip_z, flip_y)
            except IndexError as e:
                print(e)
                return

            for i, ii in enumerate(self.image_items):
                ii.setImage(slices[i], autoLevels=False, opacity=opacity)

    def set_t_threshold(self, t):
        if self.vol:
            # This takes a while, so let's have a progress indicator
            self.vol.set_t_threshold(t)

    def clear(self):
        """
        Override as we have multiple imageitems for
        :return:
        """
        self.vol = None
        # #clear the image item with an empty array
        #
        self.image_items[0].setImage(opacity=0.0)
        self.image_items[1].setImage(opacity=0.0)
        # self.item = None



from __future__ import division
from PyQt5 import  QtCore
from PyQt5.QtWidgets import QDialog
from vpv.ui.views.ui_gradient_editor import Ui_GradientEditor
import pyqtgraph as pg

class GradientEditor(QDialog):
    sigFinished = QtCore.pyqtSignal(list)  # Emits the two chosen gradients

    def __init__(self, parent):
        super(GradientEditor, self).__init__(parent.mainwindow)
        self.ui = Ui_GradientEditor()
        self.ui.setupUi(self)
        self.ui.pushButtonOk.clicked.connect(self.on_ok)
        self.pos_editor = pg.GradientWidget()
        self.ui.gridLayout.addWidget(self.pos_editor, 0, 1)
        self.neg_editor = pg.GradientWidget()
        self.ui.gridLayout.addWidget(self.neg_editor, 1, 1)

    def on_ok(self):
        pos_cm = self.pos_editor.colorMap().color
        neg_cm = list(reversed(self.neg_editor.colorMap().color))
        self.sigFinished.emit([pos_cm, neg_cm])
        self.close()


from vpv.ui.controllers import importer
from pathlib import Path
import pytest

def test_folder_filter():
    #path: str, folder_include_pattern: str
    assert importer.folder_filter('/mnt/bit_nfs/neil/mutants/output/jag2/20150409_JAG2_E14.5_13.3f_HOM_XX_REC_scaled_4.6823_pixel_14.0001/output/registrations/deformable_192_to_10/20150409_JAG2_E14.5_13.3f_HOM_XX_REC_scaled_4.6823_pixel_14.0001/20150409_JAG2_E14.5_13.3f_HOM_XX_REC_scaled_4.6823_pixel_14.0001.nrrd')


from .ImageVolume import ImageVolume
from vpv.common import read_image
import tempfile
from PIL import Image
import numpy as np


class VirtualStackVolume(ImageVolume):
    def __init__(self, *args):
        super(VirtualStackVolume, self).__init__(*args)

    def _load_data(self, file_paths, memap=True):
        """

        Parameters
        ----------
        file_paths

        Raises
        ------
        ValueError
            If an image in the stack is of the icorrect dimensions

        """
        # self.worker = LoadVirtualStackWorker(file_paths)
        # self.model.updating_started_signal.emit()
        # self.worker.progress_signal.connect(self.model.updating_msg_signal)
        # self.worker.finished.connect(self.loading_finsihed)
        # self.worker.start()

    # def loading_finsihed(self):
    #     self.model.updating_finished_signal.emit()
    #     return self.worker.memmap_result

        def sitk_load(p):
            return read_image(str(p))

        def pil_load(p):
            im = Image.open(p)
            return np.array(im)

        # SimpleITK reads in 2D bmps as 3D. So use PIL instead
        if file_paths[0].lower().endswith('.bmp'):
            reader = pil_load
        else:
            reader = sitk_load

        arr = reader(file_paths[0])
        dtype = arr.dtype
        zyx = list(arr.shape)
        zyx.insert(0, len(file_paths))
        t = tempfile.TemporaryFile()
        m = np.memmap(t, dtype=str(dtype), mode='w+', shape=tuple(zyx))
        for i, path in enumerate(sorted(file_paths)):
            img_arr = reader(path)
            m[i] = img_arr
        return m


# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'ui_main.ui'
#
# Created by: PyQt5 UI code generator 5.7
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(2000, 1000)
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout(self.centralwidget)
        self.horizontalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout_2.setSpacing(0)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.horizontalLayoutMain = QtWidgets.QHBoxLayout()
        self.horizontalLayoutMain.setObjectName("horizontalLayoutMain")
        self.horizontalLayout_2.addLayout(self.horizontalLayoutMain)
        MainWindow.setCentralWidget(self.centralwidget)
        self.toolBar = QtWidgets.QToolBar(MainWindow)
        self.toolBar.setObjectName("toolBar")
        MainWindow.addToolBar(QtCore.Qt.TopToolBarArea, self.toolBar)
        self.actionScale_bar = QtWidgets.QAction(MainWindow)
        self.actionScale_bar.setObjectName("actionScale_bar")
        self.actionOveraly_controls = QtWidgets.QAction(MainWindow)
        self.actionOveraly_controls.setObjectName("actionOveraly_controls")

        self.retranslateUi(MainWindow)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "MainWindow"))
        self.toolBar.setWindowTitle(_translate("MainWindow", "toolBar"))
        self.actionScale_bar.setText(_translate("MainWindow", "Scale bar"))
        self.actionOveraly_controls.setText(_translate("MainWindow", "Overaly controls"))




# The MIT License (MIT)
#
# Copyright (c) 2014 Mats Julian Olsen
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from inspect import isgenerator
import re
import copy


class Dict(dict):

    """
    Dict is a subclass of dict, which allows you to get AND SET(!!)
    items in the dict using the attribute syntax!

    When you previously had to write:

    my_dict = {'a': {'b': {'c': [1, 2, 3]}}}

    you can now do the same simply by:

    my_Dict = Dict()
    my_Dict.a.b.c = [1, 2, 3]

    Or for instance, if you'd like to add some additional stuff,
    where you'd with the normal dict would write

    my_dict['a']['b']['d'] = [4, 5, 6],

    you may now do the AWESOME

    my_Dict.a.b.d = [4, 5, 6]

    instead. But hey, you can always use the same syntax as a regular dict,
    however, this will not raise TypeErrors or AttributeErrors at any time
    while you try to get an item. A lot like a defaultdict.

    """

    def __init__(self, *args, **kwargs):
        """
        If we're initialized with a dict, make sure we turn all the
        subdicts into Dicts as well.

        """
        for arg in args:
            if not arg:
                continue
            elif isinstance(arg, dict):
                for key, val in arg.items():
                    self[key] = self._hook(val)
            elif isinstance(arg, tuple) and (not isinstance(arg[0], tuple)):
                self[arg[0]] = self._hook(arg[1])
            elif isinstance(arg, (list, tuple)) or isgenerator(arg):
                for key, val in arg:
                    self[key] = self._hook(val)
            else:
                raise TypeError("Dict does not understand "
                                "{0} types".format(type(arg)))

        for key, val in kwargs.items():
            self[key] = val

    def __setattr__(self, name, value):
        """
        setattr is called when the syntax a.b = 2 is used to set a value.

        """
        if hasattr(Dict, name):
            raise AttributeError("'Dict' object attribute "
                                 "'{0}' is read-only".format(name))
        else:
            self[name] = value

    def __setitem__(self, name, value):
        """
        This is called when trying to set a value of the Dict using [].
        E.g. some_instance_of_Dict['b'] = val. If 'val

        """
        super(Dict, self).__setitem__(name, value)

    @classmethod
    def _hook(cls, item):
        """
        Called to ensure that each dict-instance that are being set
        is a addict Dict. Recurses.

        """
        if isinstance(item, dict):
            return cls(item)
        elif isinstance(item, (list, tuple)):
            return type(item)(cls._hook(elem) for elem in item)
        return item

    def __getattr__(self, item):
        return self.__getitem__(item)

    def __getitem__(self, name):
        """
        This is called when the Dict is accessed by []. E.g.
        some_instance_of_Dict['a'];
        If the name is in the dict, we return it. Otherwise we set both
        the attr and item to a new instance of Dict.

        """
        if name not in self:
            self[name] = Dict()
        return super(Dict, self).__getitem__(name)

    def __delattr__(self, name):
        """ Is invoked when del some_addict.b is called. """
        del self[name]

    _re_pattern = re.compile('[a-zA-Z_][a-zA-Z0-9_]*')

    def __dir__(self):
        """
        Return a list of addict object attributes.
        This includes key names of any dict entries, filtered to the subset of
        valid attribute names (e.g. alphanumeric strings beginning with a
        letter or underscore).  Also includes attributes of parent dict class.
        """
        dict_keys = []
        for k in self.keys():
            if isinstance(k, str):
                m = self._re_pattern.match(k)
                if m:
                    dict_keys.append(m.string)

        obj_attrs = list(dir(Dict))

        return dict_keys + obj_attrs

    def _ipython_display_(self):
        print(str(self))    # pragma: no cover

    def _repr_html_(self):
        return str(self)

    def prune(self, prune_zero=False, prune_empty_list=True):
        """
        Removes all empty Dicts and falsy stuff inside the Dict.
        E.g
        >>> a = Dict()
        >>> a.b.c.d
        {}
        >>> a.a = 2
        >>> a
        {'a': 2, 'b': {'c': {'d': {}}}}
        >>> a.prune()
        >>> a
        {'a': 2}

        Set prune_zero=True to remove 0 values
        E.g
        >>> a = Dict()
        >>> a.b.c.d = 0
        >>> a.prune(prune_zero=True)
        >>> a
        {}

        Set prune_empty_list=False to have them persist
        E.g
        >>> a = Dict({'a': []})
        >>> a.prune()
        >>> a
        {}
        >>> a = Dict({'a': []})
        >>> a.prune(prune_empty_list=False)
        >>> a
        {'a': []}
        """
        for key, val in list(self.items()):
            if ((not val) and ((val != 0) or prune_zero) and
                    not isinstance(val, list)):
                del self[key]
            elif isinstance(val, Dict):
                val.prune(prune_zero, prune_empty_list)
                if not val:
                    del self[key]
            elif isinstance(val, (list, tuple)):
                new_iter = self._prune_iter(val, prune_zero, prune_empty_list)
                if (not new_iter) and prune_empty_list:
                    del self[key]
                else:
                    if isinstance(val, tuple):
                        new_iter = tuple(new_iter)
                    self[key] = new_iter

    @classmethod
    def _prune_iter(cls, some_iter, prune_zero=False, prune_empty_list=True):

        new_iter = []
        for item in some_iter:
            if item == 0 and prune_zero:
                continue
            elif isinstance(item, Dict):
                item.prune(prune_zero, prune_empty_list)
                if item:
                    new_iter.append(item)
            elif isinstance(item, (list, tuple)):
                new_item = type(item)(
                    cls._prune_iter(item, prune_zero, prune_empty_list))
                if new_item or not prune_empty_list:
                    new_iter.append(new_item)
            else:
                new_iter.append(item)
        return new_iter

    def to_dict(self):
        """ Recursively turn your addict Dicts into dicts. """
        base = {}
        for key, value in self.items():
            if isinstance(value, type(self)):
                base[key] = value.to_dict()
            elif isinstance(value, (list, tuple)):
                base[key] = type(value)(
                    item.to_dict() if isinstance(item, type(self)) else
                    item for item in value)
            else:
                base[key] = value
        return base

    def copy(self):
        """
        Return a disconnected deep copy of self. Children of type Dict, list
        and tuple are copied recursively while values that are instances of
        other mutable objects are not copied.

        """
        return Dict(self.to_dict())

    def __deepcopy__(self, memo):
        """ Return a disconnected deep copy of self. """

        y = self.__class__()
        memo[id(self)] = y
        for key, value in self.items():
            y[copy.deepcopy(key, memo)] = copy.deepcopy(value, memo)
        return y

    def update(self, d):
        """ Recursively merge d into self. """

        for k, v in d.items():
            if ((k not in self) or
                (not isinstance(self[k], dict)) or
                (not isinstance(v, dict))):
                self[k] = v
            else:
                self[k].update(v)

    def __getnewargs__(self):
        return tuple(self.items())

    def __getstate__(self):
        return self

    def __setstate__(self, state):
        self.update(state)


"""
Generate an IMPC xml for the manual annotation procedure
"""

from typing import Tuple
from lxml import etree
import yaml
from vpv.lib import addict


def get_annotator_id_and_date(procedure_id) -> Tuple[str, str]:
    """
    Get the unique annotator_id and annotation date parameter ids for a procedure.
    Parameters
    ----------
    procedure_id

    Returns
    -------
    annotator_id, date_of_annotation
    """

    map_ = {
        'IMPC_EOL': ('IMPC_EOL_052_001', 'IMPC_EOL_053_001'),
        'IMPC_EML': ('IMPC_EML_057_001', 'IMPC_EML_058_001'),
        'IMPC_EMO': ('IMPC_EMO_178_001', 'IMPC_EMO_179_001')
    }
    return map_[procedure_id[:8]]


class ExportXML(object):
    """
    TODO: project
    """
    def __init__(self,
                 date_of_annotation: str,
                 annotator_id: str,
                 metadata: str
                 ):
        """

        Parameters
        ----------
        date_of_annotation
            yyy-mm-dd
        annotator_id
            annonymized annotator id
        metadata
            path to procedure_metadata.yaml
        """

        self.metadata = load_metadata(metadata)
        md = self.metadata

        # Create separate file for each modality
        self.root = etree.Element('centreProcedureSet',
                             xmlns="http://www.mousephenotype.org/dcc/exportlibrary/datastructure/core/procedure")

        centre = etree.SubElement(self.root, 'centre', project=md['project'], pipeline=md['pipeline'],
                                  centreID=md['centre_id'])

        # Create an experiment element
        self.experiment = etree.SubElement(centre, 'experiment', dateOfExperiment=md['dateofexperiment'],
                                      experimentID=md['experiment_id'])

        # Append specimen
        specimen = etree.SubElement(self.experiment, 'specimenID')
        specimen.text = md['specimenid']

        self.procedure_element = etree.SubElement(self.experiment, 'procedure', procedureID=md['procedure_id'])

        annotator_param_id, date_of_annotation_param_id = get_annotator_id_and_date(self.metadata['procedure_id'])
        # Add metadata parameters that are not supplied in the procedure_metadata.yaml
        self.metadata['metadata'][annotator_param_id] = annotator_id
        self.metadata['metadata'][date_of_annotation_param_id] = date_of_annotation

    def add_metadata(self):
        for id_, param_value in self.metadata['metadata'].items():
            parameter = etree.SubElement(self.procedure_element, 'procedureMetadata', parameterID=id_)
            # Create value element
            value = etree.SubElement(parameter, 'value')
            value.text = str(param_value)

    def write(self, file_path):
        # print etree.tostring(root, pretty_print=True, xml_declaration=True, encoding='UTF-8', standalone='yes')
        etree.ElementTree(self.root).write(file_path, pretty_print=True, xml_declaration=True, encoding='UTF-8',
                                      standalone='yes')

    def add_series_media_parameter(self, parameter_id: str):
        """
        Add SMP which will have parameterAssociations attached to it later (annotation points)

        Parameters
        ----------
        parameter_id: the IMPC reconstruction parameter

        """
        smp = etree.SubElement(self.procedure_element,
                               'seriesMediaParameter',
                               parameterID=parameter_id)

        smp_value = etree.SubElement(smp, 'value',
                                     {"incrementValue": "1",  "URI": self.metadata['reconstruction_url']})

        # Add the deafault images parameter
        self.series_media_parameter = smp_value

    def add_point(self, param_id, xyz, xyz_percents):
        """
        Annotation points cannot be added to the manual annotation call as it is a simpleParameter
        Instead, we have to create a seriesMediaParameter and add it there with a link to the manual annotation
        parameter

        Parameters
        ----------
        param_id: str
            The parameter ID to link the point to
        param_and_id: tuple
            The annotation parameter
        xyz: tuple (int)
            the location of the annotation mark
        xyz_percents: tuple
            the percentage distances of the annotation mark from the origin

        Notes
        -----
        IMPC_EMO_001_001 is the parameter ID for embryo reconstructions


        The final series media parameter for an annotation point should look like this

            <parameterAssociation  parameterID="IMPC_EMO_022_001“>
            <dim origin="topLeft" id="x">700</dim>
            <dim origin="topLeft" id="y">350</dim>
            <dim origin="topLeft" id="z">500</dim>
            <dim origin="topLeft" id="x_percent">24.6</dim>
            <dim origin="topLeft" id="y_percent">34.7</dim>
            <dim origin="topLeft" id="z_percent">15.4</dim>
        """

        param_assoc = etree.SubElement(self.series_media_parameter,
                                       'parameterAssociation',
                                       {'parameterID': param_id})

        def put_in_points(ids, values):
            for id_, value in zip(ids, values):
                etree.SubElement(param_assoc,
                                 'dim',
                                 {'origin': 'RAS',
                                  'id': id_}).text = str(value)

        put_in_points(['x', 'y', 'z'], xyz)
        put_in_points(['x_percent', 'y_percent', 'z_percent'], xyz_percents)


    def add_parameter(self, param_id, param_value):
        """
        Add a manual annotation parameter to the XML

        Parameters
        ----------
        param_id: str
            The parameter ID for this annotation. Now ewe are using IMPRESS paarameters
            eg: IMPC_EMO_022_001  see https://www.mousephenotype.org/impress/parameterontologies/19112/495

        param_value: str
            Value associated with this annotation.
            for example: "abnormal"
        x,y,z: int or None
            the optional points associated with this annotation (can be None)
        Notes
        -----
        Adapted from James Brown's http://phobos/james.brown/export_xml_generation

        """

        # Get parameter info and append to procedure
        parameter = etree.SubElement(self.procedure_element, 'simpleParameter', parameterID=param_id)

        # Create value element
        value = etree.SubElement(parameter, 'value')
        value.text = param_value
        return parameter


def load_metadata(yaml_path):
    """
    Load in the procedure_metadata yaml file - Update the version ID to be 002 for now
    increase the procedureID if at 001 as this needs to be version 002 for validation

    Parameters
    ----------
    yaml_path: str
        path to config file

    Returns
    -------
    dict

    """

    with open(yaml_path, 'r') as fh:
        data = yaml.load(fh)

    # override the procedure ID
    proc_id: str = data['procedure_id'].replace('_001', '_002')
    data['procedure_id'] = proc_id

    return data


def load_xml(xml_file) -> Tuple:
    """
    Reads in a manual annotation xml file
    """

    try:
        root = etree.parse(xml_file)
    except (etree.XMLSyntaxError, IOError) as e:
        raise IOError(e.msg)

    root = strip_ns_prefix(root)

    simple_params = addict.Dict()
    procedure_metadata = []

    for a in root.iter():
        if a.tag == 'centre':
            centreID = a.attrib['centreID']
            pipeline = a.attrib['pipeline']
            project = a.attrib['project']

        elif a.tag == 'experiment':
            doe = a.attrib['dateOfExperiment']
            ex_id = a.attrib['experimentID']

        elif a.tag == 'specimenID':
            spec_id = a.text

        elif a.tag == 'procedure':
            # proc_id = a.attrib['procedureID']  hard code to v2
            proc_id = a.attrib['procedureID']

        elif a.tag == 'simpleParameter':
            param_id = a.attrib['parameterID']
            value = a.find('value').text
            simple_params[param_id].option = value

        elif a.tag == 'procedureMetadata':
            param_id = a.attrib['parameterID']
            value = a.find('value').text
            procedure_metadata.append((param_id, value))

        elif a.tag == 'seriesMediaParameter':
            for b in a.iter():
                param_assocs = b.findall('parameterAssociation')
                for assoc in param_assocs:
                    param_id = assoc.attrib['parameterID']
                    for dim in assoc.findall('dim'):
                        if dim.attrib['id'] == 'x':
                            x = dim.text
                        elif dim.attrib['id'] == 'y':
                            y = dim.text
                        elif dim.attrib['id'] == 'z':
                            z = dim.text
                    simple_params[param_id].xyz = (x, y, z)

    return centreID, pipeline, project, doe, ex_id, spec_id, proc_id, simple_params, procedure_metadata


def strip_ns_prefix(tree):
    """
    remove the namespace from the tree
    Parameters
    ----------
    tree

    Returns
    -------

    """
    #xpath query for selecting all element nodes in namespace
    query = "descendant-or-self::*[namespace-uri()!='']"
    #for each element returned by the above xpath query...
    for element in tree.xpath(query):
        #replace element name with its local name
        element.tag = etree.QName(element).localname
    return tree



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'ui_editor_tab.ui'
#
# Created by: PyQt5 UI code generator 5.7
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_console(object):
    def setupUi(self, console):
        console.setObjectName("console")
        console.resize(593, 603)
        self.verticalLayout = QtWidgets.QVBoxLayout(console)
        self.verticalLayout.setObjectName("verticalLayout")
        self.mainLayout = QtWidgets.QVBoxLayout()
        self.mainLayout.setObjectName("mainLayout")
        self.tableViewVolumes = QtWidgets.QTableView(console)
        self.tableViewVolumes.setObjectName("tableViewVolumes")
        self.mainLayout.addWidget(self.tableViewVolumes)
        self.verticalLayout.addLayout(self.mainLayout)
        self.pushButton = QtWidgets.QPushButton(console)
        self.pushButton.setObjectName("pushButton")
        self.verticalLayout.addWidget(self.pushButton)

        self.retranslateUi(console)
        QtCore.QMetaObject.connectSlotsByName(console)

    def retranslateUi(self, console):
        _translate = QtCore.QCoreApplication.translate
        console.setWindowTitle(_translate("console", "Form"))
        self.pushButton.setText(_translate("console", "save selected images"))

from . import resources_rc



import os
import logging
from os.path import join, dirname, abspath
from vpv.common import get_stage_and_modality, error_dialog, load_yaml, Stage, Modality
from os.path import splitext, isfile, isdir

SCRIPT_DIR = dirname(abspath(__file__))
OPTIONS_DIR = join(SCRIPT_DIR, 'options')
OPTIONS_CONFIG_PATH = os.path.join(OPTIONS_DIR, 'annotation_conf.yaml')
PROCEDURE_METADATA = 'procedure_metadata.yaml'

ANNOTATION_DONE_METADATA_FILE = '.doneList.yaml'


class Annotation(object):
    """
    Records a single manual annotation

    Attributes
    ----------
    looked_at: bool
        Whether an annotator has looked at this term and checked the done box
    """
    def __init__(self, x, y, z, dims, stage):
        self.x = None
        self.y = None
        self.z = None
        self.x_percent = None
        self.y_percent = None
        self.z_percent = None
        self.dims = dims  # x,y,z
        self.set_xyz(x, y, z)
        # self.set_xyz(x, y, z)
        self.stage = stage
        self._looked_at = False

    @property
    def looked_at(self):
        return self._looked_at

    @looked_at.setter
    def looked_at(self, done):
        self._looked_at = done

    def __getitem__(self, index):
        if index == 0:  # First row of column (dimensions)
            return "{}, {}, {}".format(self.x, self.y, self.z)  # Convert enum member to string for table
        else: # The terms and stages columns
            return self.indexes[index - 1]

    def set_xyz(self, x, y, z):
        self.x = x
        self.y = y
        self.z = z
        self.set_percentages()

    def set_percentages(self):
        dims = self.dims
        if None in (self.x, self.y, self.z):  # Annotations at startup. No position when not annotated
            return None, None, None
        self.x_percent = 100.0 / dims[0] * self.x
        self.y_percent = 100.0 / dims[1] * self.y
        self.z_percent = 100.0 / dims[2] * self.z


class ImpcAnnotation(Annotation):
    def __init__(self, x, y, z, emapa_term, name, options, default_option, dims, stage: Stage, order, is_mandatory):
        super(ImpcAnnotation, self).__init__(x, y, z, dims, stage)
        self.term = str(emapa_term)
        self.name = name
        self.options = options
        self.selected_option = default_option
        self.dims = dims
        self.order = order
        self.is_mandatory = is_mandatory
        self.indexes = [self.term, self.selected_option, self.stage]
        self.type = 'emapa'
        # self.category = category


class SpecimenAnnotations(object):
    """
    Associated with a single Volume (specimen) contains multiple Annotation
    
    For testing we're just doing the E15.5 stage. We will have to have multiple stages later and there will need
    to be some way of only allowing one stage of annotation per volume

    """

    def __init__(self, dims: tuple, vol_path: str):
        """
        Parameters
        ----------
        dims: The dimensions of the image being annotated
        vol_path: The original path of the volume being annotated
            The vol_path is used in order to look for associated procedure metadata file
        """
        self.vol_path = vol_path
        self.annotation_dir = splitext(self.vol_path)[0]
        if not isdir(self.annotation_dir):
            self.annotation_dir = None
        self.annotations = []
        self.col_count = 4  # is this needed?
        self.dims = dims
        # self.load_annotation_options() # We don't load annotations until center + stage is selected
        self.index = len(self.annotations)

        # The center where the annotation is taking place
        self.center = None
        # The developmental stage of the embryo being annotated
        self._stage = None

        self.modality = None

        # self.date_of_annotation = None  # Will be set from the annotation gui_load_done_status
        self.saved_xml_fname = None  # Will be set when xml is loaded from file

        self.annotation_date = None

        self.proc_id = None  # The imaging procedure eg IMPC_EOL__002

        self._load_options_and_metadata()
        # self._load_done_status()

    @property
    def stage(self):
        return self._stage

    @stage.setter
    def stage(self, stage):
        # After setting stage we can then set the avaialble annotiton objects
        self._stage = stage

    def _load_done_status(self):
        """
        Load in the doneList.yaml which specifies which parameters have been loked at. This is used to populate the
        'done' checkboxes, which is just a guide for the user. Has no other effect
        """
        if self.annotation_dir:
            done_file = join(self.annotation_dir, ANNOTATION_DONE_METADATA_FILE)
            if not isfile(done_file):
                return

            done_status = load_yaml(done_file)
            if not done_status:
                return

            self.index = len(self.annotations)  # bodge. Need to reset the annotations iterator

            for ann in self:
                done = done_status.get(ann.term, 'notpresent')

                if done is 'notpresent':
                    logging.warning('Cannot find term in done metadata\n{}'.format(done_file))

                else:
                    ann.looked_at = done

    def _load_options_and_metadata(self):
        """
        The volume has been loaded. Now see if there is an associated annotation folder that will contain the IMPC
        metadata parameter file. Also load in any partially completed xml annotation files.

        """
        if not self.annotation_dir:
            return

        self.metadata_parameter_file = join(self.annotation_dir, PROCEDURE_METADATA)
        
        if not isfile(self.metadata_parameter_file):
            self.metadata_parameter_file = None
            return

        cso = centre_stage_options.opts

        metadata_params = load_yaml(self.metadata_parameter_file)
        if not metadata_params:
            return

        proc_id = metadata_params['procedure_id']

        # Temp fix 030320 to increment procedures to 002 so they vallidate correctly
        proc_id = proc_id.replace('_001', '_002')
        self.proc_id = proc_id

        center_id = metadata_params['centre_id']
        stage_id, modality = get_stage_and_modality(proc_id, center_id)

        self.stage = stage_id
        self.center = center_id
        self.modality = modality

        # Get the procedure parameters for the given center/stage
        try:
            cso['centers'][center_id]['procedures'][proc_id]
        except KeyError:
            logging.error(f'Procedure id {proc_id} not in the annotation_conf')
            return

        center_stage_default_params = cso['centers'][center_id]['procedures'][proc_id]['parameters']

        # opts['centers'][center]['procedures'][proc_id]['parameters'] = self.load_centre_stage_file(param_file_)

        for _, param_info in center_stage_default_params.items():

            try:
                options = centre_stage_options.opts['available_options'][param_info['options']]
            except TypeError as e:
                logging.error('Falied to load annotation parameter file {}'.format(e))
                return

            if param_info['default_option'] not in options:

                logging.error(
                    """Annotation parameter list load error
                    default option: {} not in available options:{}
                    Centre:, stage:{}""".format(param_info['default_option'], options, center_id, stage_id))

                error_dialog(None, 'Annotation options load error', "See log file - 'info/show log'")
                return


            default = param_info['default_option']

            self.add_impc_annotation(None,
                                     None,
                                     None,
                                     param_info['impc_id'],
                                     param_info['name'],
                                     options,
                                     default,
                                     self.stage,
                                     param_info['order'],
                                     param_info['mandatory'],
                                     self.dims
                                     )
        # Sort the list and set the interator index
        self.annotations.sort(key=lambda x: x.order, reverse=True)
        self.index = len(self.annotations)

    def update_annotation(self, term, x, y, z, selected_option: str):
        """
        Update an annotation
        Returns
        -------

        """
        ann = self.get_by_term(term)
        ann.set_xyz(x, y, z)
        # ann.x, ann.y, ann.z = x, y, z
        ann.selected_option = selected_option

    def add_impc_annotation(self, x: int, y:int, z:int, impc_param, name, options, default_option, stage: Stage, order,
                            is_mandatory, dims):
        """
        Add an emap type annotation from available terms on file
        """
        ann = ImpcAnnotation(x, y, z, impc_param, name, options, default_option, dims, stage, order, is_mandatory)
        self.annotations.append(ann)

    def remove(self, row):
        del self.annotations[row]

    def clear(self):
        self.annotations = []
        self.index = 0

    def __getitem__(self, index):
        return self.annotations[index]

    def __len__(self):
        return len(self.annotations)

    def __iter__(self):
        return self

    def __next__(self):
        if self.index == 0:
            self.index = len(self.annotations)
            raise StopIteration
        self.index -= 1
        return self.annotations[self.index]

    def get_by_term(self, term: str) -> Annotation:
        for annotation in self:
            if annotation.term == term:
                self.index = len(self.annotations)  # Bodge. How am I supposed to reset index
                return annotation
        self.index = len(self.annotations)
        return False


class CenterStageOptions(object):
    def __init__(self):
        # Add assert for default in options
        try:
            opts = load_yaml(OPTIONS_CONFIG_PATH)
        except OSError:
            logging.warning('could not open the annotations yaml file {}'.format(OPTIONS_CONFIG_PATH))
            raise
        self.opts = opts
        self.available_options = self.opts['available_options']

        # Get the centre/stage-specific parameters

        for centre_id, centre_data in opts['centers'].items():
            for proc_id, proc_data in centre_data['procedures'].items():
                param_file_ = proc_data['file_']

                # now add the parameters from each individual centre/stage file to the main options config
                try:
                    opts['centers'][centre_id]['procedures'][proc_id]['parameters'] = self.load_centre_stage_file(param_file_)
                except KeyError as e:
                    raise

    def load_centre_stage_file(self, yaml_name):
        """
        Load in centrer-specific annotation parameter list
        Parameters
        ----------
        yaml_name: str
            basename of annotation parameter file

        Returns
        -------
        dict


        """
        path = join(OPTIONS_DIR, yaml_name)
        opts = load_yaml(path)

        # Rename the param_1, param_2 keys with IMPC parameter key
        renamed = {}

        for k in list(opts['parameters']):

            new_key = opts['parameters'][k]['impc_id']
            renamed[new_key] = opts['parameters'][k]

        opts['parameters'] = renamed
        return opts['parameters']

    def all_stages(self, center):
        return [x for x in self.opts['centers'][center]['stages'].keys()]

    def all_centers(self):
        return [x for x in self.opts['centers']]

centre_stage_options = CenterStageOptions()



from PyQt5 import QtGui
import pyqtgraph as pg
from vpv.common import Orientation
import numpy as np
import math


class VectorLayer(object):
    def __init__(self, viewbox, parent, model):
        # Create dummy vectors
        self.model = model
        self.parent = parent
        self.orientation = self.parent.orientation
        self.slice_change_function = None
        self.arrows = []
        self.viewbox = viewbox
        self.def_field = None
        self.drawn = False
        self.item = None
        self.arrow_angle = 50
        self.plt = pg.PlotItem()
        self.plt.hideAxis('left')
        self.plt.hideAxis('right')
        self.plt.hideAxis('bottom')
        self.plt.hideAxis('top')
        self.vol = None
        self.arrow_color = "22E300"
        self.vec_mag_min = 0.0
        self.vec_mag_max = 5.0

    def set_magnitude_cutoff(self, min_, max_):
        self.vec_mag_min = min_
        self.vec_mag_max = max_
        self.set_slice(self.current_index)

    def register_slice_change_function(self):
        if self.orientation == Orientation.coronal:
            return self.vol.get_coronal
        elif self.orientation == Orientation.sagittal:
            return self.vol.get_sagittal
        elif self.orientation == Orientation.axial:
            return self.vol.get_axial

    def set_volume(self, volname):
        if volname == "None":
            self.volume_label_signal.emit("None")
            self.vol = None
            if self.item:
                self.viewbox.removeItem(self.item)
            return
        self.volume_label_signal.emit(volname)

        self.parent.overlay.set_vector_label(volname)

        self.vol = self.model.getvol(volname)
        orientation = self.parent.orientation
        self.slice_change_function = self.register_slice_change_function()
        self.name = self.vol.name
        dim_len = self.vol.dimension_length(orientation)
        self.current_index = dim_len / 2
        midslice = self.slice_change_function(dim_len / 2)
        self.shape = midslice.shape

        self.plt.hideAxis('left')
        self.plt.hideAxis('right')
        self.plt.hideAxis('bottom')
        self.plt.hideAxis('top')
        self.plt.hideButtons()
        self.plt.setRange(xRange=(0, self.shape[0]), yRange=(0, self.shape[1]))
        # self.plt.yRange(shape[2])
        self.viewbox.addItem(self.plt)

        self.set_orientation()

        self.set_slice(dim_len / 2)

    def set_orientation(self):
        """
        set which 2 axes to take from the 3D vector.
        """
        if self.parent.orientation == Orientation.axial:
            self.vector_axes = (1, 0)
        elif self.parent.orientation == Orientation.coronal:
            self.vector_axes = (0, 2)
        elif self.parent.orientation == Orientation.sagittal:
            self.vector_axes = (1, 2)

    def set_slice(self, index=None, flip=None):
        # The flip has not been tested for vector display and is probably broke
        # if not self.vol:
        #     if self.item:
        #         self.viewbox.removeItem(self.item)
        #     print 'not vol'
        #     return

        if index < 0:
            return
        if not index:
            index = self.current_index
        else:
            self.current_index = index - 1

        c = self.vol.subsampling
        scale = self.vol.scale

        if self.item:
            self.viewbox.removeItem(self.item)

        # Get the slice and 2D vectors for this orientation
        slice_ = self.slice_change_function(index)

        # Get the 2d vector for this plane
        slice_2d_vec = slice_.take(self.vector_axes, axis=2)

        x_points = []
        y_points = []
        connect = []

        for y in range(0, slice_.shape[0] - c, c):
            for x in range(0, slice_.shape[1] -c, c):

                try:

                    subsmapled_box = slice_2d_vec[y: y+c, x: x+c, :]
                except IndexError:
                    print("fell off")
                else:
                    x1 = x + (c/2)
                    y1 = y + (c/2)

                    x_magnitude, y_magnitude = np.mean(subsmapled_box, axis=(0, 1))
                    magnitude = np.linalg.norm((x_magnitude, y_magnitude))
                    if magnitude < self.vec_mag_min:
                        continue
                    if self.orientation == Orientation.axial:
                        x_magnitude, y_magnitude = self.rotate_vector((x_magnitude, y_magnitude), -90.0)

                    x_points.append(x1)
                    y_points.append(y1)
                    connect.append(1)

                    # Draw a line to end of arrow
                    x2 = x1 + (x_magnitude * scale)
                    y2 = y1 + (y_magnitude * scale)

                    x_points.append(x2)
                    y_points.append(y2)
                    connect.append(1)

                    arrow_Xs, arrow_ys = self.draw_arrow_head(x2, x1, y2, y1)

                    x_points.append(arrow_Xs[0])
                    y_points.append(arrow_ys[0])
                    connect.append(0)

                    x_points.append(arrow_Xs[1])
                    y_points.append(arrow_ys[1])
                    connect.append(1)

                    # Back to arrow tip
                    x_points.append(x2)
                    y_points.append(y2)
                    connect.append(0)

        path = pg.arrayToQPath(np.array(x_points), np.array(y_points), np.array(connect))
        self.item = QtGui.QGraphicsPathItem(path)
        self.item.setPen(pg.mkPen({'color': self.arrow_color, 'width': 1}))
        self.viewbox.addItem(self.item)

    def rotate_vector(self, vector, theta):
        x_temp = np.copy(vector[0])
        theta = math.radians(theta)
        x = vector[0] * math.cos(theta) - vector[1] * math.sin(theta)
        y = x_temp * math.sin(theta) + vector[1] * math.cos(theta)
        return x, y

    def draw_arrow_head(self, tipX, tailX, tipY, tailY):
        """
        This is the function that needs speeding up
        :param tipX:
        :param tailX:
        :param tipY:
        :param tailY:
        :return:
        """
        arrowLength = 1
        dx = tipX - tailX
        dy = tipY - tailY

        theta = math.atan2(dy, dx)

        rad = math.radians(20)
        x = tipX - arrowLength * math.cos(theta + rad)
        y = tipY - arrowLength * math.sin(theta + rad)


        phi2 = math.radians(-20)
        x2 = tipX - arrowLength * math.cos(theta + phi2)
        y2 = tipY - arrowLength * math.sin(theta + phi2)

        arrowYs = []
        arrowYs.append(y)
        arrowYs.append(y2)

        arrowXs = []
        arrowXs.append(x)
        arrowXs.append(x2)

        return arrowXs, arrowYs

    def set_subsampling(self, value):
        self.vol.subsampling = int(value)
        self.update()

    def set_scale(self, value):
        self.vol.scale = int(value)
        self.update()

    def set_arrow_color(self, color):
        self.arrow_color = color

    def clear(self):
        pass

    def update(self):
        self.set_orientation()
        self.set_slice(self.current_index)

    def reload(self):
        pass




from PyQt5 import QtGui, QtCore
from vpv.ui.views.ui_options_tab import Ui_options
from vpv.utils.appdata import AppData


"""
A widget that allows the setting of some global options. The widget is added as a tab on the data manager
tab widget
"""


class OptionsTab(QtGui.QWidget):
    flip_signal = QtCore.pyqtSignal()
    filter_label_signal = QtCore.pyqtSignal(list)

    def __init__(self, mainwindow, appdata: AppData):
        super(OptionsTab, self).__init__(mainwindow)
        self.ui = Ui_options()
        self.ui.setupUi(self)

        self.ui.checkBoxAxialFlipx.stateChanged.connect(self.on_axial_flipx)
        self.ui.checkBoxAxialFlipY.stateChanged.connect(self.on_axial_flipy)
        self.ui.checkBoxAxialFlipZ.stateChanged.connect(self.on_axial_flipz)

        self.ui.checkBoxCoronalFlipX.stateChanged.connect(self.on_coronal_flipx)
        self.ui.checkBoxCoronalFlipY.stateChanged.connect(self.on_coronal_flipy)
        self.ui.checkBoxCoronalFlipZ.stateChanged.connect(self.on_coronal_flipz)

        self.ui.checkBoxSagittalFlipX.stateChanged.connect(self.on_sagittal_flipx)
        self.ui.checkBoxSagittalFlipY.stateChanged.connect(self.on_sagittal_flipy)
        self.ui.checkBoxSagittalFlipZ.stateChanged.connect(self.on_sagittal_flipz)

        self.all_flip_check_boxes = [
            self.ui.checkBoxAxialFlipx,
            self.ui.checkBoxAxialFlipY,
            self.ui.checkBoxAxialFlipZ,
            self.ui.checkBoxCoronalFlipX,
            self.ui.checkBoxCoronalFlipY,
            self.ui.checkBoxCoronalFlipZ,
            self.ui.checkBoxSagittalFlipX,
            self.ui.checkBoxSagittalFlipY,
            self.ui.checkBoxSagittalFlipZ
        ]
        self.impc_flip_boxes = [
            self.ui.checkBoxAxialFlipZ,
            self.ui.checkBoxAxialFlipY,
            self.ui.checkBoxCoronalFlipZ
        ]

        self.ui.checkBoxImpcView.clicked.connect(self.on_impc_view)

        self.appdata = appdata

        # 080419 testing out filtering by label volumes on layer 2
        self.ui.lineEditShowLabel.textChanged.connect(self.filter_label)

    def filter_label(self):
        input_ = self.ui.lineEditShowLabel.text()
        labels = input_.split(' ')
        try:
            labels = [int(x) for x in labels]
        except (TypeError, ValueError):
            labels = [0] # Reset the filtering
        self.filter_label_signal.emit(labels)

    def set_orientations(self):
        """
        Load the orientation-specific flip state stored in the appdata
        """
        if not self.appdata:
            return

        self.flips = self.appdata.get_flips()

        axial= self.flips['axial']
        coronal = self.flips['coronal']
        sagittal = self.flips['sagittal']

        self.ui.checkBoxAxialFlipx.setChecked(axial['x'])

        self.ui.checkBoxAxialFlipY.setChecked(axial['y'])

        self.ui.checkBoxAxialFlipZ.setChecked(axial['z'])

        self.ui.checkBoxCoronalFlipX.setChecked(coronal['x'])

        self.ui.checkBoxCoronalFlipY.setChecked(coronal['y'])

        self.ui.checkBoxCoronalFlipZ.setChecked(coronal['z'])

        self.ui.checkBoxSagittalFlipX.setChecked(sagittal['x'])

        self.ui.checkBoxSagittalFlipY.setChecked(sagittal['y'])

        self.ui.checkBoxSagittalFlipZ.setChecked(sagittal['z'])

        self.ui.checkBoxImpcView.setChecked(self.flips['impc_view'])

        self.on_impc_view(self.flips['impc_view'])

    def on_impc_view(self, checked: bool):
        """
        IMPC view is a series of flips to make the data look like as agreed.
        Uncheck all and set agred on flips
        """
        self.flips['impc_view'] = checked
        if checked:
            for box in self.all_flip_check_boxes:
                if box not in self.impc_flip_boxes:
                    box.setEnabled(False)
                    box.setChecked(False)
                else:
                    box.setEnabled(True)
                    box.setChecked(True)

        else:
            for box in self.all_flip_check_boxes:
                box.setEnabled(True)

        self.flip_signal.emit()

    def on_axial_flipx(self, checked: int):
        self.flips['axial']['x'] = bool(checked)
        self.flip_signal.emit()

    def on_axial_flipy(self, checked: int):
        self.flips['axial']['y'] = bool(checked)
        self.flip_signal.emit()

    def on_axial_flipz(self, checked: int):
        self.flips['axial']['z'] = bool(checked)
        self.flip_signal.emit()

    def on_coronal_flipx(self, checked: int):
        self.flips['coronal']['x'] = bool(checked)
        self.flip_signal.emit()

    def on_coronal_flipy(self, checked: int):
        self.flips['coronal']['y'] = bool(checked)
        self.flip_signal.emit()

    def on_coronal_flipz(self, checked: int):
        self.flips['coronal']['z'] = bool(checked)
        self.flip_signal.emit()

    def on_sagittal_flipx(self, checked: int):
        self.flips['sagittal']['x'] = bool(checked)
        self.flip_signal.emit()

    def on_sagittal_flipy(self, checked: int):
        self.flips['sagittal']['y'] = bool(checked)
        self.flip_signal.emit()

    def on_sagittal_flipz(self, checked: int):
        self.flips['sagittal']['z'] = bool(checked)
        self.flip_signal.emit()




"""
This script loads a series of volumes and associated labels from a config file
And displays them according to some options that can be set.

Currently works with the data structure as output by lama, but will add option to specify paths

Examples
--------

Example yaml config file



TODO:
Set opacity and label maps per view

"""
import sys
from pathlib import Path
from itertools import chain

import yaml
from PyQt5 import QtGui

from vpv.vpv_temp import Vpv
from vpv.common import Layers, Orientation
import addict


def load(loader_file):

    with open(loader_file, 'r') as fh:
        config = addict.Dict(yaml.load(fh))

    top = config['top_row']
    top_vols = [Path(x.vol.path) for x in top.values()]
    top_labels = [Path(x.labels.path) for x in top.values()]

    bottom = config['bottom_row']
    if bottom: # We allow only top tier visible
        bottom_vols = [Path(x.vol.path) for x in bottom.values()]
        bottom_labels = [Path(x.labels.path) for x in bottom.values()]
    else:
        bottom_vols = []
        bottom_labels = []

    app = QtGui.QApplication([])
    ex = Vpv()

    s_ = lambda x: [str(z) for z in x]

    ex.load_volumes(chain(s_(top_vols), s_(top_labels), s_(bottom_vols), s_(bottom_labels)), 'vol')

    # Vpv deals with images with the same name by appending parenthetical digits. We need to know the ids it will assign
    # if we are to get a handle once loaded
    img_ids = ex.img_ids()
    num_top_views = len(top_vols)
    num_bottom_views = len(bottom_vols)

    # Set the top row of views
    counter = 0
    for i in range(num_top_views):
        try:
            # vol_id = top_vols[i].stem
            vol_id = img_ids[counter]
            # label_id = top_labels[i].stem
            label_id = img_ids[counter + len(top_vols)]
            # if label_id == vol_id:
            #     label_id = f'{label_id}(1)'
            ex.views[i].layers[Layers.vol1].set_volume(vol_id)
            ex.views[i].layers[Layers.vol2].set_volume(label_id)
            counter += 1
        except IndexError:
            continue

    if bottom:
        # Set the bottom row of views
        for i in range(num_bottom_views):
            try:
                # vol_id = top_vols[i].stem
                vol_id = img_ids[counter + num_bottom_views]
                # label_id = top_labels[i].stem
                label_id = img_ids[counter + num_bottom_views *2]
                # if label_id ==vol_id:
                #     label_id = f'{label_id}(1)'
                ex.views[i + 3].layers[Layers.vol1].set_volume(vol_id)
                ex.views[i + 3].layers[Layers.vol2].set_volume(label_id)
                counter += 1
            except IndexError:
                continue

    print('Finished loading')

    # Show two rows
    ex.data_manager.show2Rows(True if bottom else False)

    # Set orientation
    ex.data_manager.on_orientation('sagittal')

    # Set colormap
    ex.data_manager.on_vol2_lut_changed('anatomy_labels')

    # opacity
    ex.data_manager.modify_layer(Layers.vol2, 'set_opacity', config.opacity)

    sys.exit(app.exec_())


if __name__ == '__main__':
    import sys
    load(sys.argv[1])


import numpy as np
import math
import SimpleITK as sitk

test_vector_out = '/home/neil/share/deformations_test/test_vector_def.nrrd'
test_vector_out_mhd = '/home/neil/share/deformations_test/test_vector_def.mhd'

def rotate_vector(vector, theta):
    theta = math.radians(theta)
    x_temp = float(vector[0])
    x = vector[0] * math.cos(theta) - vector[1] * math.sin(theta)
    y = x_temp * math.sin(theta) + vector[1] * math.cos(theta)
    return x, y

a = np.zeros(100*100*100*3).reshape((100, 100, 100, 3))
# Make a vector field of 2
a.fill(2)

# For each slice, rotate 45 degress
theta = -90
for y in range(a.shape[1]):
    theta += 90
    if theta > 360:
        theta = 0
    for z in range(a.shape[0]):
        for x in range(a.shape[2]):
            rotated = rotate_vector(a[z,y,x, [0,1]], theta)
            a[z, y, x,[0,1]] = rotated

out = sitk.GetImageFromArray(a)
sitk.WriteImage(out, test_vector_out)
sitk.WriteImage(out, test_vector_out_mhd)









#!/usr/bin/env python3

# Copyright 2016 Medical Research Council Harwell.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# @author Neil Horner <n.horner@har.mrc.ac.uk>


# I distribute VPV as an installer with a bundled version of WinPython
# In at least one case, the paths are not correctly set. So the following hack attempts to set correct paths


WINPYTHON_DIR = 'WinPython-64bit-3.6.3.0Zero'
PYTHON_DIR = 'python-3.6.3.amd64'

import os
import sys
from pathlib import Path
import logging
from os.path import join, isdir
from typing import Iterable, List
p = sys.path

from PyQt5 import QtGui, QtCore, QtWidgets
from vpv import common
from vpv.ui.controllers import importer

from vpv.ui.controllers import log_viewer

try:
    QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling, True)
    QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_UseHighDpiPixmaps, True)
except AttributeError:
    logging.info("High DPI scaling not available. QT >=5.6 is needed for this ")

if os.name == 'nt':
    # check where vpv_viewer has been installed
    vpv_installation_dir = os.path.dirname(os.path.realpath(__file__))
    winpython_path = os.path.join(vpv_installation_dir, WINPYTHON_DIR, PYTHON_DIR)
    if os.path.isdir(winpython_path):
        dll_path = os.path.join(winpython_path, 'DLLs')
        lib_path = os.path.join(winpython_path, 'Lib')
        site_packages_path = os.path.join(lib_path, 'site-packages')
        sys.path.insert(0, dll_path)
        sys.path.insert(0, lib_path)
        sys.path.insert(0, site_packages_path)
    else:
        logging.info('cannot find winpython folder: {}'.format(winpython_path))

from vpv.ui.controllers.dock_widget_manager import ManagerDockWidget
from vpv.model.model import DataModel
from vpv.utils.appdata import AppData
from vpv.common import Orientation, Layers, log_path, error_dialog
from vpv.display.slice_view_widget import SliceWidget
from vpv.ui.controllers.data_manager import ManageData
from vpv.ui.controllers.options_tab import OptionsTab
from vpv.annotations.annotations_widget import AnnotationsWidget
from vpv.model.coordinate_mapper import Coordinate_mapper
from vpv.ui.controllers import main_window
from vpv.utils import github


try:
    from vpv.ui.controllers.console import Console
    console_imported = True
except ImportError as e:
    logging.info('cannot import qtconsole, so diabling console widget tab\n{e}')
    console_imported = False
except Exception:  # I thnk it might not be an ImportError? look into it
    logging.info('cannot import qtconsole, so diabling console widget tab')
    console_imported = False



from vpv.ui.controllers.gradient_editor import GradientEditor
import zipfile
from vpv.lib import addict
import tempfile
import csv
import logging.config
from vpv.common import log_path



class Vpv(QtCore.QObject):
    """The entry point to VPV. Acts a bit like a controller taking signals from the different views and propogating
    them to other views.
    """
    data_processing_finished_signal = QtCore.pyqtSignal()
    crosshair_visible_signal = QtCore.pyqtSignal()
    crosshair_invisible_signal = QtCore.pyqtSignal()
    volume_pixel_signal = QtCore.pyqtSignal(float)
    volume2_pixel_signal = QtCore.pyqtSignal(float)
    heatmap_pixel_signal = QtCore.pyqtSignal(float)
    # volume_position_signal = QtCore.pyqtSignal(int, int, int)

    def __init__(self):
        super(Vpv, self).__init__()

        self.voxel_size = 28.0
        self.view_scale_basrs = False
        self.view_id_counter = 0
        self.appdata = AppData()

        print(self.appdata.monster_munch)
        self.mainwindow = main_window.Mainwindow(self, self.appdata)
        # self.mainwindow.showFullScreen()
        self.model = DataModel()
        self.model.updating_started_signal.connect(self.updating_started)
        self.model.updating_finished_signal.connect(self.updating_finished)
        self.model.updating_msg_signal.connect(self.display_update_msg)
        self.views = {}
        # display and views now created in manage_views
        self.data_manager = ManageData(self, self.model, self.mainwindow, self.appdata)
        self.data_manager.gradient_editor_signal.connect(self.gradient_editor)

        self.annotations_manager = AnnotationsWidget(self, self.mainwindow)
        self.annotations_manager.annotation_highlight_signal.connect(self.show_saved_annotations)
        self.annotations_manager.annotation_radius_signal.connect(self.annotation_radius_changed)
        self.annotations_manager.roi_highlight_off_signal.connect(self.reset_roi)

        self.mainwindow.d_pressed_signal.connect(self.annotations_manager.d_pressed_slot)
        self.mainwindow.key_up_down_signal.connect(self.annotations_manager.sroll_annotations)

        self.volume_pixel_signal.connect(self.mainwindow.set_volume_pix_intensity)
        self.volume2_pixel_signal.connect(self.mainwindow.set_volume2_pix_intensity)
        self.heatmap_pixel_signal.connect(self.mainwindow.set_data_pix_intensity)

        self.options_tab = OptionsTab(self.mainwindow, self.appdata)
        self.options_tab.flip_signal.connect(self.update_slice_views)

        self.options_tab.filter_label_signal.connect(self.filter_label)

        # Sometimes QT console is a pain to install. If not availale do not make console tab
        if console_imported:
            self.console = Console(self.mainwindow, self)
        else:
            self.console = None

        self.dock_widget = ManagerDockWidget(self.model, self.mainwindow, self.appdata, self.data_manager,
                                             self.annotations_manager, self.options_tab, self.console)
        self.dock_widget.setAllowedAreas(QtCore.Qt.LeftDockWidgetArea)

        # Create the initial 3 orthogonal views. plus 3 hidden for the second row
        inital_views = [
            [Orientation.sagittal, 'red', 0, 1, False],
            [Orientation.coronal, 'blue', 0, 2, False],
            [Orientation.axial, 'green', 0, 3, False],
            [Orientation.sagittal, 'orange', 1, 1, True],
            [Orientation.coronal, 'grey', 1, 2, True],
            [Orientation.axial, 'cyan', 1, 3, True]
        ]

        self.mapper = Coordinate_mapper(self.views, self.appdata.get_flips())

        for v in inital_views:
            self.setup_views(*v)
        self.current_view = self.views[0]

        self.crosshair_visible = False

        self.data_manager.roi_signal.connect(self.mapper.roi_to_view)

        self.any_data_loaded = False
        self.crosshair_permanent = False
        self.gradient_editor_widget = None

        self.options_tab.set_orientations()

        self.annotation_radius_changed(self.appdata.annotation_circle_radius)

        self.check_vpv_version()

    def filter_label(self, labels: Iterable[int]):
        """
        Recieves a label to show. Pass to slice views

        Parameters
        ----------
        labels
            the label numbers to show, if 0 set to view all

        """
        for v in self.views.values():
            v.filter_label(labels)

    def set_orientation_visibility(self, visible: bool):
        for view in self.views.values():
            view.set_orientation_labels_visiblility(visible)

    def check_vpv_version(self):
        """
        Check the mpi2 github page for new versions
        If newer version available, set the link in a label on the data manager tab
        """
        # Catch all exceptions as we don't want this to cause problems for the rest of the app
        try:
            new_version = github.get_latest_github_version()
        except Exception:
            return

        if new_version:
            self.data_manager.ui.labelNewVersion.setText('New verion available\n{}'.format(new_version))

    def show_log(self):
        """
        Show the log in a scrossable window
        """

        log_viewer.Logview(self.mainwindow, log_path)

    def on_slice_view_mouse_move(self, x: int, y: int, z: int, src_view: SliceWidget):
        """
        Given coordinates of mouse hover position, emit signals to update the voxel value indicators.
        If shift key is pressed, activate synchronised viewing

        Parameters
        ----------
        src_view
            the emitting slice view widge
        x, y
            the hover coordinates in the 2D slice view
        z
            The current slice index of the slice view
        """

        vol = src_view.main_volume
        vol2 = src_view.secondary_volume
        hm = src_view.heatmap_volume

        if not vol:
            return

        if any(i < 0 for i in (x, y, z)):
            return

        # map to the volume space
        vol_points = self.mapper.view_to_volume(x, y, z, src_view.orientation, src_view.main_volume.shape_xyz())

        x1 = src_view.main_volume.shape_xyz()[0] - vol_points[0]

        self.mainwindow.set_mouse_position_indicator(*vol_points)

        # Get any flips that have been applied to the data
        flipx, flipy, flipz = src_view.get_flips()
        # Get the values of the voxels underneath the mouse pointer
        try:
            vol_hover_voxel_value = vol.get_data(Orientation.axial, vol_points[2],
                                                 xy=[x1, vol_points[1]])

        except IndexError:
            pass
        else:
            if x > 0 and x > 0:
                self.volume_pixel_signal.emit(round(float(vol_hover_voxel_value), 2))
                if vol2:
                    vol2_hover_voxel_value = vol2.get_data(Orientation.axial, vol_points[2], xy=[x1, vol_points[1]])
                    self.volume2_pixel_signal.emit((round(float(vol2_hover_voxel_value), 4)))
                if hm:
                    hm_hover_voxel_value = hm.get_data(Orientation.axial, vol_points[2], xy=[x1, vol_points[1]])
                    self.heatmap_pixel_signal.emit((round(float(hm_hover_voxel_value), 4)))

        # # If shift is pressed emit signal to get other views to get to the same or interscting slice
        modifiers = QtGui.QApplication.keyboardModifiers()

        if modifiers == QtCore.Qt.ShiftModifier:

            # With mouse move signal, also send current vol.
            # If veiews are not synchronised, syncyed slicing only occurs within the same volumes
            self.mouse_shift(x, y, z, src_view)

    def map_annotation_signal_view_to_view(self, slice_idx: int, x: int, y: int, src_view: SliceWidget,
                                           color: tuple=(255, 0, 0), radius: int=10, reverse: bool=False):
        """
        Upon getting a mouse click on a SliceWidget region, it will emit info here including position and the
        emitting view.  Map the coordinates between views so that they are correctly positioned

        Parameters
        ----------
        slice_idx: the current slice index of the emitting view
        x: the y position of the view clicked
        y: the y position of the view clicked
        src_view: the emitting view
        color: rgb color of the annotation marker
        radius: radius of the annotation marker
        reverse:

        """
        if not self.current_annotation_volume():
            self.annotations_manager.reset_roi()
            return

        if not self.annotations_manager.annotating:
            return
        dims = self.current_annotation_volume().shape_xyz()
        for dest_view in self.views.values():
            # First map the annotation marker between views
            dest_x, dest_y, dest_index = self.mapper.view_to_view(x, y, slice_idx, src_view.orientation,
                                                                  dest_view.orientation, dims)

            dest_view.set_slice(dest_index)
            # Set the annotation marker. Red for pre-annoation, green indicates annotation save
            dest_view.show_annotation_marker(dest_x, dest_y, color)

            # Add the coordinates to the AnnotationsWidget
            # The coordinates need to be in axial space do map back to them if necessary
            if dest_view.orientation == Orientation.axial:
                # Now map it back to image coordinates
                xa, ya, idxa = self.mapper.view_to_volume(x, y, slice_idx, src_view.orientation, dims)
                self.annotations_manager.set_annotation_position_label(xa, ya, idxa)

    def mouse_shift(self, x, y, z, src_view):
        """
        Gets mouse moved signal. Sets corresponding slices in other views if linked

        Parameters
        ----------
        z: int
            the current slice of the calling view
        x: int
            the x position of the mouse
        y: int
            the y position of the mouse
        src_view: SliceWidget
            the calling view
        """

        if not self.current_annotation_volume():
            return

        dims = self.current_annotation_volume().shape_xyz()

        for dest_view in self.views.values():

            dest_x, dest_y, dest_z = self.mapper.view_to_view(x, y, z, src_view.orientation, dest_view.orientation, dims)

            try:
                dest_view.set_slice(dest_z, crosshair_xy=(dest_x, dest_y))
            except IndexError:
                pass

    def current_annotation_volume(self):
        """
        Get the volume currently visible in the annotation view
        
        Returns 
        -------
        model.ImageVolume instance
        """
        if not self.current_view:
            return None
        return self.current_view.layers[Layers.vol1].vol

    def on_console_enter_pressesd(self):
        self.update_slice_views()

    def take_screen_shot(self):
        # Hide the sliders, take screenshot, then show slider
        sshot = self.mainwindow.ui.centralwidget.grab()
        QtGui.QApplication.clipboard().setPixmap(sshot)

        if common.question_dialog(self.mainwindow, "Screenshot copied to clipboard", 'Save to file?'):

            path = QtWidgets.QFileDialog.getSaveFileName(self.mainwindow, 'Save screen shot',
                                                  self.appdata.last_screen_shot_dir)
            print(path)
            if path[0]:
                self.appdata.last_screen_shot_dir = str(Path(path[0]).parent)
                sshot.save(path[0])

        # common.info_dialog(self.mainwindow, 'Message', "Screenshot copied to clipboard")

    def volume_changed(self, vol_name):
        self.data_manager.volume_changed(vol_name)

    def set_current_view(self, slice_id):
        self.current_view = self.views[slice_id]

    def reset_roi(self):
        for view in self.views.values():
            view.switch_off_annotation()

    def current_orientation(self):
        return self.current_view.orientation

    def annotation_radius_changed(self, radius):
        for view in self.views.values():
            view.annotation_radius_changed(radius)

    def tab_changed(self, indx: int):
        """
        When changing to annotations tab, make sure all views are linked
        """
        self.data_manager.link_views = True
        self.annotations_manager.tab_changed(indx)

    def updating_started(self):
        self.updating_dlg = QtGui.QMessageBox()

    def updating_finished(self):
        self.updating_dlg.close()

    def display_update_msg(self, msg: str):
        self.updating_dlg.setText(msg)

    def set_view_controls_visibility(self, visible):
        for view in self.views.values():
            view.show_index_slider(visible)

    def show_color_scale_bars(self, visible):

        self.data_manager.show_color_scale_bars(visible)

    def show_scale_bars(self, visible):
        for view in self.views.values():
            view.update()
            view.show_scale_bar(visible)

    def update_slice_views(self):

        for view in self.views.values():
            view.update_view()

    def setup_views(self, orientation: Orientation, color: tuple, row: int, column: int, hidden: bool=False):
        """
        Create all the orthogonal views and setup the signals and slots
        """
        view = self.add_view(self.view_id_counter, orientation, color, self.mapper)
        # view.mouse_shift.connect(self.mouse_shift)
        # view.mouse_pressed_signal.connect(self.dock_widget.mouse_pressed)
        view.mouse_pressed_annotation_signal.connect(self.map_annotation_signal_view_to_view)
        # view.mouse_pressed_annotation_signal.connect(self.annotations_manager.mouse_pressed_annotate)
        view.crosshair_visible_signal.connect(self.crosshair_visible_slot)
        view.scale_changed_signal.connect(self.zoom_changed)
        view.slice_index_changed_signal.connect(self.index_changed)
        view.move_to_next_vol_signal.connect(self.move_to_next_vol)
        self.data_manager.scale_bar_color_signal.connect(view.set_scalebar_color)
        self.crosshair_visible_signal.connect(view.show_crosshair)
        self.crosshair_invisible_signal.connect(view.hide_crosshair)
        self.view_id_counter += 1
        self.mainwindow.add_slice_view(view, row, column)
        view.setHidden(hidden)

        # Connect the mouse moving so we can get pixel/label value and position
        view.mouse_moved_signal.connect(self.on_slice_view_mouse_move)

    def gradient_editor(self):
        # Activeated 6 times on one click so bogdge for now
        self.ge = GradientEditor(self)
        self.ge.sigFinished.connect(self.set_heatmap_luts)
        self.ge.show()

    def set_heatmap_luts(self, luts):
        for view in self.views.values():
            if view.layers[Layers.heatmap].vol:
                view.layers[Layers.heatmap].vol.pos_lut = luts[0]
                #view.display[Layer.heatmap].vol.neg_lut = luts[1]
                view.layers[Layers.heatmap].update()

    def move_to_next_vol(self, view_id, reverse=False):
        if self.data_manager.link_views:
            for view in self.views.values():
                view.move_to_next_volume(reverse)
        else:
            self.views[view_id].move_to_next_volume(reverse)
        self.data_manager.switch_selected_view(view_id)

    def recalc_connected_components(self):
        self.current_view.layers[Layers.heatmap].vol.find_largest_connected_components()
        self.data_manager.update_connected_components(self.current_view.layers[Layers.heatmap].vol.name)

    def add_view(self, id_, orientation, color, mapper):
        """
        Setup each orthogonal viewer

        Parameters
        ----------
        id_: int
        orientation: Orientation
        Initial orientation
        color: str
            just a color word at the moment eg: red
        mapper coordinate_mapper.Coordinate_mapper
            Hold info such as whether the current orientation should have flips applied

        Returns
        -------
        SliceWidget

        """
        view = SliceWidget(orientation, self.model, color, mapper)
        view.manage_views_signal.connect(self.update_manager)
        self.views[id_] = view
        return view

    def toggle_dock_widget_visibility(self):
        if self.dock_widget.isVisible():
            self.dock_widget.hide()
        else:
            self.update_manager()

    def toggle_link_views(self):
        if self.data_manager.link_views:
            link = False
        else:
            link = True
        self.data_manager.on_link_views(link)

    def update_manager(self, slice_id=0):  # called from data_manager:update_manager
        """
        update the slice manager with data from a slice view
        :param slice_id: Id of the SliceWidget that this current widget was activated from
        """
        self.current_view = self.views[slice_id]
        self.data_manager.switch_selected_view(slice_id)
        self.mainwindow.show_manager(self.dock_widget)
        self.mainwindow.addDockWidget(QtCore.Qt.LeftDockWidgetArea, self.dock_widget)
        self.dock_widget.show()

    def remove_views(self, row, column):
        self.mainwindow.remove_view(row, column)

    def index_changed(self, orientation, id_, idx):
        orientation = str(orientation)
        if not self.data_manager.link_views:
            return
        for view in self.views.values():
            if view.id == id_:  # The view that emiited the signal
                continue
            if orientation == view.orientation: # For now, only zoom views with the same orientation
                view.set_slice(idx)

    def zoom_changed(self, orientation, id_, ranges):
        """
        Called when a slice view zooms in.
        Parameters
        ----------
        orientation: string
            orientation of the calling slice view
        id: int
            id of the calling slice view
        scale: list
            [[xstart, xend], [ystart, yend]]
        """
        if not self.data_manager.link_views:
            return
        for view in self.views.values():
            if view.id == id_:  # The view that emiited the signal
                continue
            if orientation == view.orientation: # For now, only zoom views with the same orientation
                view.set_zoom(ranges[0], ranges[1])

    def crosshair_visible_slot(self, visible):
        if visible:
            self.crosshair_visible_signal.emit()
        elif not self.crosshair_permanent:
            self.crosshair_invisible_signal.emit()

    def set_crosshair_permanent(self, is_visible_always):
        self.crosshair_permanent = is_visible_always
        self.crosshair_invisible_signal.emit()

    def data_labels_visible(self, visible):
        for view in self.views.values():
            view.set_data_label_visibility(visible)

    def data_processing_finished_slot(self):
        self.data_processing_finished_signal.emit()

    def show_saved_annotations(self, x: int, y: int, z: int, color: list, radius: int):
        """
        When clicking on a saved annotation  from the table, highlight in 'color' the location
        Bit of a bodge this. Need to move some of it to coordinate mapper


        Parameters
        ----------
        x
        y
        z
        color
        radius

        Returns
        -------

        """
        dims = self.current_annotation_volume().shape_xyz()
        for dest_view in self.views.values():
            dest_x, dest_y, dest_z = self.mapper.view_to_view(x, y, z, Orientation.axial, dest_view.orientation,
                                                              dims, from_saved=True)
            dest_view.set_slice(dest_z)
            dest_view.show_annotation_marker(dest_x, dest_y, color)

    def control_visiblity(self, visible):
        for slice_ in self.slice_widgets.values():
            slice_.show_controls(visible)

    def scale_bar_visible(self, visible):
        pass
        for slice_ in self.slice_widgets.values():
            slice_.scale_bar_visible(visible)

    def load_data_slot(self, dragged_files=None):
        """

        Parameters
        ----------
        dragged_files

        Returns
        -------

        """
        last_dir = self.appdata.get_last_dir_browsed()
        # Convert QStrings to unicode in case they contain special characters
        files = [x for x in dragged_files]
        importer.Import(self.mainwindow, self.importer_callback, self.virtual_stack_callback, last_dir, self.appdata,
                        files)

    def check_non_ras(self):
        return
        for vol in self.model.all_volumes():
            if vol.dir_ != common.RAS_DIRECTIONS:
                common.info_dialog(self.mainwindow, "Image format warning!",
                                   "At lease one of your loaded volumes is not in RAS format For IMPC annotation enure HARP v2.3.0 or aboave. Or use download data from the DCC")
                break

    def browse_files(self):
        last_dir = self.appdata.get_last_dir_browsed()
        if last_dir:
            last_dir = last_dir[0]
            if not os.path.isdir(last_dir):
                files = QtWidgets.QFileDialog.getOpenFileNames(self.mainwindow, "Select files to load")
            else:
                files = QtWidgets.QFileDialog.getOpenFileNames(self.mainwindow, "Select files to load", last_dir)
        self.load_data_slot(files[0])

    def clear_views(self):
        self.model.clear_data()
        for view in self.views.values():
            view.clear_layers()
        self.data_manager.clear()
        self.annotations_manager.clear()
        self.any_data_loaded = False
        # self.update()

    def virtual_stack_callback(self, file_paths, last_dir):
        if len(file_paths) > 0:
            self.appdata.set_last_dir_browsed(last_dir)
            try:
                self.model.add_volume(file_paths, 'virtual_stack', memory_map=True)
            except ValueError:
                QtGui.QMessageBox.warning(self.mainwindow, 'Loading error',
                    "Virtual stack could not be loaded\nAre all images the same dimension?", QtGui.QMessageBox.Ok)
            else:
                self.appdata.set_last_dir_browsed(last_dir)
                if self.dock_widget.isVisible():
                    self.dock_widget.update()
                if not self.any_data_loaded:
                    #  Load up one of the volumes just loaded into the bottom layer
                    self.add_initial_volume()
                    self.any_data_loaded = True

    def add_initial_volume(self):
        """
        Called when loading in volumes for the first time, to get a volume displayed without havong to manually set it
        """
        try:
            init_vol = self.model.volume_id_list()[0]
            for view in self.views.values():
                view.layers[Layers.vol1].set_volume(init_vol, initial=True)
                # view.display[Layer.vol1].update()  # This should make sure 16bit images are scaled correctly at loading?
                view.update_view()
        except IndexError:  # No Volume objects have been loaded
            logging.warn('No volumes loaded')

        try:  # See if we have any Data objects loaded
            init_vol = self.model.data_id_list()[0]
            for view in self.views.values():
                view.layers[Layers.heatmap].set_volume(init_vol, initial=True)
        except IndexError:  # No Volume objects have been loaded
            pass

    def importer_callback(self, volumes, heatmaps, annotations, vector_files, image_series,
                          impc_analysis, last_dir, memory_map=False, distribute:bool=False):
        """
        Recieves a list of files to open from the Importer widget

        Parameters
        ----------
        dual_datafiles: str
            path to config file for loading paired tval/pval files
        :return:
        """
        if len(impc_analysis) > 0:
            self.load_impc_analysis(impc_analysis[0])
        if len(volumes) > 0:
            self.load_volumes(volumes, 'vol', memory_map)
        if len(heatmaps) > 0:
            self.load_volumes(heatmaps, 'heatmap', memory_map, fdr_thresholds=False)
        if len(vector_files) > 0:
            self.load_volumes(vector_files, 'vector', memory_map)
        if len(image_series) > 0:
            self.model.load_image_series(image_series, memory_map)
            if not self.any_data_loaded:
                self.add_initial_volume()

        # Now load the annotations from the Importer. Only load if there is a corresponding volume with the same id
        if len(annotations) > 0:
            self.load_annotations(annotations) # From importer

        self.appdata.set_last_dir_browsed(last_dir)
        self._auto_load_annotations(volumes)

        if self.dock_widget.isVisible():
            self.data_manager.update()
            self.annotations_manager.update()

        if distribute:
            self.distribute_volumes_across_views()

    def distribute_volumes_across_views(self):
        """
        When loading a volume, the same volume is loaded into into Slice view .
        This method tries to distribute amongs the views and set sets them to all the same orientation

        """
        # Get the available volumes
        vol_list = self.model.volume_id_list()

        if len(vol_list) < 2:
            return

        else:
            for i, view in self.views.items():
                view.layers[Layers.vol1].set_volume(vol_list[i])
                view.set_orientation(Orientation.sagittal)
                if len(vol_list) < 4:
                    if i >= 2:
                        return
                else:
                    if i >= 5:
                        self.data_manager.show2Rows(True)
                        return

    def _auto_load_annotations(self, volumes):
        """
        Look for previously-written annotation files in  the annotation directory (folder with same name as loaded img)
        If present try to load
        """
        import fnmatch

        annotation_xml_files = []

        for vol_path in volumes:

            # The annotation dir path will be the same as the image minus the file extension
            ann_dir = os.path.splitext(vol_path)[0]

            if not ann_dir:
                return

            if not isdir(ann_dir):
                logging.info("{}, is not a valid annotation directory".format(ann_dir))
                continue

            ann_files = os.listdir(ann_dir)
            xml_annotation_file = None

            for af in ann_files:
                if fnmatch.fnmatch(af, '*experiment.impc.xml'):
                    xml_annotation_file = join(ann_dir, af)
                    break

            if xml_annotation_file:
                annotation_xml_files.append(xml_annotation_file)

        if annotation_xml_files:
            self.load_annotations(annotation_xml_files)

    def load_annotations(self, annotation_file_list):
        """
        Load annotations that were previously saved as xml

        Parameters
        ----------
        annotation_file_list

        Returns
        -------

        """
        non_loaded = []
        for path in annotation_file_list:  # At this point looked_at set to True
            error = self.model.load_annotation(path)
            if error:
                non_loaded.append(path)
                common.error_dialog(self.mainwindow, 'Annotations not loaded', error)
            else:
                # Load annotations - these should be called within self.model.load_annotation
                self.annotations_manager.populate_available_terms()
                self.annotations_manager.update()
        if not non_loaded:
            common.info_dialog(self.mainwindow, 'Load success', 'Annotations loaded')

    def load_volumes(self, file_list, data_type, memory_map=False, fdr_thresholds=False):
        """
        Load volumes from a list of paths.

        Parameters
        ----------
        file_list: list
            list of paths to volumes
        data_type: str
            vol data etc
        memory_map: bool
            whether to memory map after reading
        fdr_thresholds: dict
            q -> t statistic mappings
                {0.01: 3.4,
                0.05:, 3.1}
        """

        non_loaded = []

        for i, vol_path in enumerate(file_list):
            try:

                self.model.add_volume(vol_path, data_type, memory_map, fdr_thresholds)
                self.appdata.add_used_volume(vol_path)
                if not self.any_data_loaded:
                    #  Load up one of the volumes just loaded into the bottom layer
                    self.add_initial_volume()
                    self.any_data_loaded = True
            except (IOError, RuntimeError) as e:  # RT error Raised by SimpleITK
                print(e)
                non_loaded.append(vol_path)
        if len(non_loaded) > 0:
            common.error_dialog(self.mainwindow, 'Volumes not loaded', '\n'.join(non_loaded))
        # self.any_data_loaded
        self.check_non_ras()
        self._auto_load_annotations(file_list)

    def img_ids(self):
        return self.model.volume_id_list(sort=False)

    def load_impc_analysis(self, impc_zip_file):
        """
        Load a zip file containing the results of the IMPC automated analysis (TCP pipeline)
        Parameters
        ----------
        impc_zip_file: str
            path tho analysis results zip
        """
        try:
            zf = zipfile.ZipFile(impc_zip_file)
        except zipfile.BadZipFile as e:
            common.error_dialog(self.mainwindow, 'Zip file error', 'Cannot read IMPC analysis zip file\nThe zip may be corrupted')
            return
        names = zf.namelist()

        file_names = addict.Dict(
            {'intensity_tstats_file': None,
             'jacobians_tstats_file': None,
             'qvals_intensity_file': None,
             'qvals_jacobians_file': None,
             'popavg_file': None}
        )

        files_remaining = []
        for name in names:
            name_lc = name.lower()
            if 'intensities-tstats' in name_lc:
                file_names.intensity_tstats_file = name
            elif 'jacobians-tstats' in name_lc:
                file_names.jacobians_tstats_file = name
            elif 'qvals-intensities' in name_lc:
                file_names.qvals_intensity_file = name
            elif 'qvals-jacobians' in name_lc:
                file_names.qvals_jacobians_file = name
            elif 'popavg' in name_lc:
                file_names.popavg_file = name
            else:
                files_remaining.append(name)

        if all(file_names.values()):
            td = tempfile.TemporaryDirectory()
            zf.extractall(td.name)
            popavg = join(td.name, file_names.popavg_file)
            self.load_volumes([popavg], 'vol')

            # get the trhesholds from the csv files
            qval_int_csv = join(td.name, file_names.qvals_intensity_file)
            intensity_fdr_thresh = self.extract_fdr_thresholds(qval_int_csv)

            inten_tstat = join(td.name, file_names.intensity_tstats_file)

            self.load_volumes([inten_tstat], 'heatmap', memory_map=False,
                              fdr_thresholds=intensity_fdr_thresh)

            qval_jac_csv = join(td.name, file_names.qvals_jacobians_file)
            jacobian_fdr_thresh = self.extract_fdr_thresholds(qval_jac_csv)

            jac_tstat = join(td.name, file_names.jacobians_tstats_file)

            self.load_volumes([jac_tstat], 'heatmap', memory_map=False,
                              fdr_thresholds=jacobian_fdr_thresh)

            # Load any other volumes in the zip. Probably will be mutants
            mutants = [join(td.name, x) for x in files_remaining if x.endswith('nrrd')]
            self.load_volumes(mutants, 'vol', memory_map=False)

            if not intensity_fdr_thresh:
                common.info_dialog(self.mainwindow, "No hits",
                                   "There are no hits in the intensity heatmap. The threshold is set to max")
            if not jacobian_fdr_thresh:
                common.info_dialog(self.mainwindow, "No hits",
                                   "There are no hits in the jacobian heatmap. The threshold is set to max")
        else:
            failed = []
            for key, name in file_names.items():
                if not name:
                    failed.append(key)
            common.error_dialog(self.mainwindow, 'Error loading file', 'The following files could not be found in the zip\n {}'.format('\n'.join(failed)))
            print('IMPC analysis data failed to load. The following files could not be found in the zip')
            print(failed)

    @staticmethod
    def extract_fdr_thresholds(stats_info_csv):
        """
        Given a csv path containing the stats summary from the TCP pipeline (or LAMA)
        read the fdr threshold q value and corresponding t-statsitc into a dict

        Parameters
        ----------
        stats_info_csv: str
            path to csv
        Returns
        -------
        dict of q to t mappings
            {0.1: 2.6,
            0.2: 2.2...}
        None if there are no t-thresholds for any given q-value cutoff
        """
        from collections import OrderedDict
        q_t = OrderedDict()

        with open(stats_info_csv, 'r') as csvfile:
            reader = csv.reader(csvfile, delimiter=',')
            reader.__next__()  # remove header

            for line in reader:

                try:
                    q = float(line[0])
                except ValueError:
                    print("There is a problem with the stats info file. Cannot read the q-value '{}'from file {}".format(
                        line[0], stats_info_csv
                    ))
                    return {}  # return an empty dict
                try:
                    t = float(line[3])
                except ValueError:  # NAs are here when there are below the minimu, threshold
                    t = None
                q_t[q] = t
        if any(q_t.values()):
            return q_t
        else:
            return None  # No t thresholds. Proably no hits in this heatmap

    def activate_view_manager(self, view_widget_id):
        """
        :param sliceid: int ,
        :return:
        """
        self.dock_widget.show(view_widget_id)

    def close(self):
        print('saving settings to {}'.format(self.appdata.app_data_file))
        self.appdata.write_app_data()
        self.model.write_temporary_annotations_metadata()
        print('exiting')

    def on_view_new_screen(self):

        # Work out how to maximize on anotehr screen
        # dtw = QtGui.QDesktopWidget()
        # current_screen = dtw.screenNumber(self.mainwindow)
        # print 'scree', dtw.screenGeometry(3)

        pass
        # self.mainwindow2 = main_window.Main(self, self.app_data)
        # self.mainwindow2.setWindowTitle('Harwell vol viewer (2)')
        #
        # self.slice_sag2 = SliceWidget('sagittal', self.model, 'yellow')
        # self.slice_axi2 = SliceWidget('axial', self.model, 'magenta')
        # self.slice_cor2 = SliceWidget('coronal', self.model, 'orange')
        #
        # self.slice_cor2.manage_views_signal.connect(self.activate_view_manager)
        # self.slice_axi2.manage_views_signal.connect(self.activate_view_manager)
        # self.slice_sag2.manage_views_signal.connect(self.activate_view_manager)
        #
        # self.slice_widgets2 = [self.slice_sag2, self.slice_cor2, self.slice_axi2]
        #
        # self.mainwindow2.add_slice_views(self.slice_widgets2)
        # self.model.register_slice_widgets(self.slice_widgets2)


def update_check():
    pass





# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'ui_layer_widget.ui'
#
# Created by: PyQt5 UI code generator 5.7
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_Layer(object):
    def setupUi(self, Layer):
        Layer.setObjectName("Layer")
        Layer.resize(1030, 51)
        self.horizontalLayout = QtWidgets.QHBoxLayout(Layer)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setObjectName("gridLayout")
        self.comboBoxVol = QtWidgets.QComboBox(Layer)
        self.comboBoxVol.setObjectName("comboBoxVol")
        self.gridLayout.addWidget(self.comboBoxVol, 0, 0, 1, 1)
        self.comboBoxLut = QtWidgets.QComboBox(Layer)
        self.comboBoxLut.setObjectName("comboBoxLut")
        self.gridLayout.addWidget(self.comboBoxLut, 0, 1, 1, 1)
        self.horizontalLayout.addLayout(self.gridLayout)

        self.retranslateUi(Layer)
        QtCore.QMetaObject.connectSlotsByName(Layer)

    def retranslateUi(self, Layer):
        _translate = QtCore.QCoreApplication.translate
        Layer.setWindowTitle(_translate("Layer", "Form"))




# Copyright 2016 Medical Research Council Harwell.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# @author Neil Horner <n.horner@har.mrc.ac.uk>
import re

from PyQt5 import QtGui, QtCore
from PyQt5.QtWidgets import QDialog, QMessageBox, QFileDialog
import os
from vpv.ui.views.ui_importer import Ui_Dialog

VOLUME = 'Volume'
HEATMAP = 'Heatmap data'
LAMA_DATA = 'LAMA data'
VECTORS = "Vectors"
ANNOTATIONS = "Annotations"
IMAGE_SERIES = "Image series"
IMPC_ANALYSIS = "IMPC analysis"

TYPE_CHOICES = (VOLUME, HEATMAP, ANNOTATIONS, VECTORS, IMAGE_SERIES, IMPC_ANALYSIS)


class Import(QDialog):
    def __init__(self, parent, callback, virtual_stack_callback, last_dir, appdata, dragged_files=None):
        super(Import, self).__init__(parent)
        self.ui = Ui_Dialog()
        self.ui.setupUi(self)
        self.last_dir = last_dir
        self.callback = callback
        self.appdata = appdata
        self.virtual_stack_callback = virtual_stack_callback

        # Connect up the buttons for movin gthe rows around
        # self.ui.pushButtonUp.clicked.connect(self.move_up)
        # self.ui.pushButtonDown.clicked.connect(self.move_down)

        # Setup signals
        self.ui.pushButtonCancel.clicked.connect(self.on_cancel)
        self.ui.pushButtonChooseFile.clicked.connect(self.on_choose_files)
        self.ui.pushButtonChoseDir.clicked.connect(self.on_load_virtual_stack)
        self.ui.pushButtonOK.clicked.connect(self.on_ok)
        self.ui.checkBoxAllSameType.clicked.connect(self.on_all_same_type)
        self.type_combo_boxes = None
        self.ui.virtualStackWidget.hide()
        self.ui.pushButtonFilter.clicked.connect(self.filter_virtual_stack)

        # Setup folder filtering
        self.ui.lineEditFolderFilter.editingFinished.connect(self.set_folder_filter_pattern)

        # Create the table
        self.files_to_open = []
        self.folder_include_pattern = None

        self.vs_file_list = []
        self.vs_file_list_to_ignore = set()

        self.use_virtual_stack = False
        self.all_same_type = False

        if dragged_files:
            self.files_to_open = dragged_files
            # If one path is dropped onto the main window and it's a directory, as if we want to load a stack of slices
            if len(dragged_files) == 1 and os.path.isdir(dragged_files[0]):

                # use_folder_of_slices = QtGui.QMessageBox.question(parent, 'Choose data type?',
                #                                                   "Load individual 2D slices from folder (yes)\n"
                #                                                   "Load multiple image volumes (No)",
                #                                                   'Hello' | QtGui.QMessageBox.No)
                box = QMessageBox()
                box.setIcon(QMessageBox.Question)
                box.setWindowTitle('Choose data type')
                box.setText('Load individual 2D slices from folder, or load multiple single images volumes?')
                box.setStandardButtons(QMessageBox.Yes | QMessageBox.No)
                button_y = box.button(QMessageBox.Yes)
                button_y.setText('Load stack from individual 2D image slices')
                button_n = box.button(QMessageBox.No)
                button_n.setText('Load multiple 3D image volumes')
                box.exec_()
                if box.clickedButton() == button_y:
                    self.setWindowTitle("Load stack of 2D slices")
                    self.load_virtual_stack(dragged_files[0])
                else:
                    self.populate_file_list()
                    self.setWindowTitle("Load image volumes")
            else:
                self.populate_file_list()
                self.setWindowTitle("Load image volumes")
        else:
            self.populate_file_list()
            self.setWindowTitle("Load image volumes")

        self.show()

    def set_folder_filter_pattern(self):
        """
        Gets a signal to indicate folder filtering has been specified.
        Gets the filter pattern an filters the available images accordingly
        """
        pattern = self.ui.lineEditFolderFilter.text()
        self.folder_include_pattern = str(pattern)
        self.populate_file_list()

    def folder_filter(self, path: str) -> bool:
        """

        Returns
        -------
        True if path matches pattern
        False if path does not match pattern
        """
        # path

        if not self.folder_include_pattern:
            return True  # Do no filtering as there's no pattern

        # Replace any asterix in the pattern to '.+' to match anything` ````
        pattern = self.folder_include_pattern.replace('*', '.+')

        # 211019 temp bodge: ignore these subfolders that keep getting in the way
        if 'resolution_images' in path:
            return False

        if re.search(pattern, path, re.IGNORECASE):
            return True  # Found pattern in path
        else:
            return False

    def filter_virtual_stack(self):
        contains_text = self.ui.lineEditFileNameContains.displayText()
        contains = [x.strip() for x in contains_text.split(',') if x]
        dn_contain_text = self.ui.lineEditFileNameDoesNotContain.displayText()
        dn_contain = [x.strip() for x in dn_contain_text.split(',') if x]
        files_to_use = set()
        files_to_remove = set()
        for c in contains:
            files_to_use.update([x for x in self.vs_file_list if c in x])
        for d in dn_contain:
            files_to_remove.update([x for x in self.vs_file_list if d in x])
        self.vs_file_list_to_ignore.update(set(self.vs_file_list).difference(files_to_use))
        self.vs_file_list_to_ignore.update(files_to_remove)
        self.populate_table_for_virtual_stack()

    def on_all_same_type(self, checked):
        self.all_same_type = checked

    def move_down(self):
        row = self.ui.table.currentRow()
        column = self.ui.table.currentColumn()
        if row < self.ui.table.rowCount() - 1:
            self.ui.table.insertRow(row + 2)
            for i in range(self.ui.table.columnCount()):
                self.ui.table.setItem(row + 2, i, self.ui.table.takeItem(row, i))
                #self.ui.table.setCurrentCell(row + 2, i)
            self.ui.table.removeRow(row)

    def move_up(self):
        row = self.ui.table.currentRow()
        column = self.ui.table.currentColumn();
        if row > 0:
            self.ui.table.insertRow(row - 1)
            for i in range(self.ui.table.columnCount()):
                self.ui.table.setItem(row - 1, i, self.ui.table.takeItem(row + 1, i))
                self.ui.table.setCurrentCell(row - 1, column)
            self.ui.table.removeRow(row + 1)

    def on_cancel(self):

        self.close()

    def on_choose_files(self):

        qstring_filenames = QFileDialog.getOpenFileNames(self, "Choose some data", self.last_dir)
        self.files_to_open = [x for x in qstring_filenames]
        self.populate_file_list()

    def set_last_dir(self, dir_):
        self.last_dir = dir_

    def on_load_virtual_stack(self):
        dir_ = QFileDialog.getExistingDirectory(None, "Selet a folder of of images to create virtual stack")
        self.load_virtual_stack(dir_)

    def load_virtual_stack(self, dir_):
        """
        Load a series of images into a virtual stack
        :return:
        """
        self.ui.virtualStackWidget.show()
        # Load the saved fitler terms
        saved_include_patterns = self.appdata.get_include_filter_patterns()
        saved_exclude_patterns = self.appdata.get_exclude_filter_patterns()

        last_dir = dir_
        self.last_dir = last_dir
        self.vs_file_list = self.get_all_filles_in_dir(dir_)
        self.use_virtual_stack = True
        self.populate_table_for_virtual_stack()
        self.ui.lineEditFileNameContains.setText("{}".format(','.join(saved_include_patterns)))
        self.ui.lineEditFileNameDoesNotContain.setText("{}".format(','.join(saved_exclude_patterns)))

    def get_files_for_virtual_stack(self):
        file_list = []
        for row in range(self.ui.table.rowCount()):
            if self.ui.table.cellWidget(row, 2).isChecked():
                path = self.ui.table.item(row, 0).data(QtCore.Qt.UserRole)
                file_list.append(path)
        self.virtual_stack_callback(file_list, self.last_dir)
        self.ui.virtualStackWidget.hide()
        self.close()

    def on_ok(self):
        """
        Load the data from the table that has not been unchecked
        :return:
        """
        if self.use_virtual_stack:
            self.get_files_for_virtual_stack()
            # Save the file filter patterns
            exclude_patterns = self.ui.lineEditFileNameDoesNotContain.text().strip().split(',')
            include_patterns = self.ui.lineEditFileNameContains.text().strip().split(',')
            self.appdata.set_include_filter_patterns(include_patterns)
            self.appdata.set_exclude_filter_patterns(exclude_patterns)
            return
        volumes = []
        datafiles = []
        vector_files = []
        image_series = []
        annotations = []
        impc_analysis = []

        for row in range(self.ui.table.rowCount()):
            if self.ui.table.cellWidget(row, 2).isChecked():
                type_ = str(self.ui.table.cellWidget(row, 1).currentText())
                if type_ == VOLUME:
                    vol = self.ui.table.item(row, 0).data(QtCore.Qt.UserRole)
                    volumes.append(vol)
                elif type_ == HEATMAP:
                    datafiles.append(self.ui.table.item(row, 0).data(QtCore.Qt.UserRole))
                elif type_ == VECTORS:
                    vector_files.append(self.ui.table.item(row, 0).data(QtCore.Qt.UserRole))
                elif type_ == IMAGE_SERIES:
                    image_series.append(self.ui.table.item(row, 0).data(QtCore.Qt.UserRole))
                elif type_ == ANNOTATIONS:
                    annotations.append(self.ui.table.item(row, 0).data(QtCore.Qt.UserRole))
                elif type_ == IMPC_ANALYSIS:
                    impc_analysis.append(self.ui.table.item(row, 0).data(QtCore.Qt.UserRole))

        self.callback(volumes, datafiles, annotations, vector_files, image_series, impc_analysis,
        self.last_dir, self.ui.checkBoxMemoryMap.isChecked(), self.ui.checkBoxDistrbute.isChecked())

        self.close()

    def get_all_filles_in_dir(self, dir_):

        files_to_open = []
        for path, subdirs, files in os.walk(dir_):
            for name in files:
                path_ = os.path.join(path, name)
                if self.guess_type(path_):
                    files_to_open.append(os.path.join(path, name))

        # Now check if there's a stats yaml file in there
        # stats_config = os.path.join(dir_, 'stats.yaml')
        # if os.path.isfile(stats_config):
        #     with open(stats_config) as yf:
        #         config = yaml.load(yf)
        #         target = config.get('target')
        #         if target:
        #             script_dir = os.path.dirname(os.path.abspath(__file__))
        #             target_path = os.path.abspath(os.path.join(script_dir, target))
        #             files_to_open.append(target_path)
        return files_to_open

    def type_combo_changed(self, index):
        if self.all_same_type:
            if self.type_combo_boxes:
                for cb in self.type_combo_boxes:
                    cb.setCurrentIndex(index)

    def populate_file_list(self):

        # Clear table
        self.ui.table.setRowCount(0)

        files_to_open = self.files_to_open

        folder_paths = []
        folders_to_remove = []

        # If a single directory has been dropped onto vpv_viewer
        for path in files_to_open:
            if os.path.isdir(path):
                folder_paths.extend(self.get_all_filles_in_dir(path))
                folders_to_remove.append(path)

        if len(folder_paths) > 0:
            files_to_open.extend(folder_paths)

        for folder in folders_to_remove:
            if folder in files_to_open:
                files_to_open.remove(folder)

        # Filter based on pattern
        files_to_open = [x for x in files_to_open if self.folder_filter(x)]

        self.ui.table.setRowCount(len(files_to_open))

        self.type_combo_boxes = []

        for row, item in enumerate(files_to_open):

            type_guess = self.guess_type(item)

            if not self.folder_filter(item):
                continue

            if not type_guess:
                continue

            data_path = QtGui.QTableWidgetItem(item)
            data_path.setText(os.path.basename(item))
            data_path.setData(QtCore.Qt.UserRole, item)
            self.ui.table.setItem(row, 0, data_path)

            type_combo = QtGui.QComboBox()
            type_combo.addItems(TYPE_CHOICES)
            type_combo.setCurrentIndex(type_combo.findText(type_guess))
            type_combo.activated.connect(self.type_combo_changed)
            self.type_combo_boxes.append(type_combo)
            self.ui.table.setCellWidget(row, 1, type_combo)

            load_checkbox = QtGui.QCheckBox()
            load_checkbox.setChecked(True)
            self.ui.table.setCellWidget(row, 2, load_checkbox)

        self.ui.table.resizeColumnsToContents()
        table_width = self.ui.table.horizontalHeader().length()
        self.setFixedWidth(table_width + 43)
        # As a bodge to get the last browsed dir, get the path of the first file in the list
        try:
            self.set_last_dir(os.path.split(files_to_open[0]))[0]
        except:
            pass

    def populate_table_for_virtual_stack(self):
        filelist = sorted(set(self.vs_file_list).difference(self.vs_file_list_to_ignore))
        self.ui.labelNumberOfFiles.setText(str(len(filelist)))
        self.ui.table.clear()
        self.ui.table.setRowCount(len(filelist))

        for row, item in enumerate(filelist):

            data_path = QtGui.QTableWidgetItem(item)
            data_path.setText(os.path.basename(item))
            data_path.setData(QtCore.Qt.UserRole, item)
            self.ui.table.setItem(row, 0, data_path)

            load_checkbox = QtGui.QCheckBox()
            load_checkbox.setChecked(True)
            self.ui.table.setCellWidget(row, 2, load_checkbox)

        self.ui.table.resizeColumnsToContents()
        table_width = self.ui.table.horizontalHeader().length()
        self.setFixedWidth(table_width + 43)

    def guess_type(self, data_path):
        volume_types = ['.nrrd', '.tiff', '.tif', '.nii', '.mnc', '.npz', '.bmp', '.json', '.gz', '.zip']
        extension = os.path.splitext(data_path)[1].lower()

        data_basname = os.path.basename(data_path)

        if extension == '.zip':
            return IMPC_ANALYSIS

        if extension == '.xml':
            return ANNOTATIONS

        if extension not in volume_types:
            return False

        heatmap_suggestions = ('stats', 'inten', 'jac')
        if any(x in data_basname for x in heatmap_suggestions):
            return HEATMAP
        if 'vector' in data_basname:
            return VECTORS

        else:
            return VOLUME

    def get_files_from_dir(self, dir_):

        self.populate_file_list([dir_])



# Copyright 2016 Medical Research Council Harwell.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
# @author Neil Horner <n.horner@har.mrc.ac.uk>

"""
TODO: don't duplicate the full array for each _get_* function
"""
import numpy as np
import os
import tempfile
from PIL import Image
from PyQt5 import QtCore
from vpv.common import read_image, get_stage_and_modality, error_dialog
from vpv.annotations.impc_xml import load_xml, get_annotator_id_and_date
from vpv.annotations.annotations_model import centre_stage_options, PROCEDURE_METADATA, ANNOTATION_DONE_METADATA_FILE

from .ImageVolume import ImageVolume
from .HeatmapVolume import HeatmapVolume
from .VectorVolume import VectorVolume
from .ImageSeriesVolume import ImageSeriesVolume
from .VirtualStackVolume import VirtualStackVolume
import yaml


class LoadVirtualStackWorker(QtCore.QThread):
    progress_signal = QtCore.pyqtSignal([str])

    def __init__(self, file_paths):
        QtCore.QThread.__init__(self)
        self.file_paths = file_paths
        self.memmap_result = None  # Populated at end of run()

    def sitk_load(self, p):
        read_image(p, convert_to_ras=True)

    def pil_load(self, p):
        im = Image.open(p)
        return np.array(im)

    def run(self):
        size = len(self.file_paths)
        # SimpleITK reads in 2D bmps as 3D. So use PIL instead
        if self.file_paths[0].lower().endswith('.bmp'):
            reader = self.pil_load
        else:
            reader = self.sitk_load

        arr = reader(self.file_paths[0])
        dtype = arr.dtype
        zyx = list(arr.shape)
        zyx.insert(0, len(self.file_paths))
        t = tempfile.TemporaryFile()
        m = np.memmap(t, dtype=dtype, mode='w+', shape=tuple(zyx))
        for i, path in enumerate(sorted(self.file_paths)):
            img_arr = reader(path)
            m[i] = img_arr
            self.progress_signal.emit("Loading virtual stack.. {}%".format(str(100.0/size * i)))
        self.memmap_result = m


class DataModel(QtCore.QObject):
    """
    The model for our app
    """
    data_changed_signal = QtCore.pyqtSignal()
    updating_started_signal = QtCore.pyqtSignal()
    updating_msg_signal = QtCore.pyqtSignal(str)
    updating_finished_signal = QtCore.pyqtSignal()

    def update_msg_slot(self, msg):
        """
        Gets update messages from the different volume classes which are then propagated to the main window to display
        a progress message

        Parameters
        ----------
        msg: str
            progress message
        """
        self.update_msg_signal.emit(msg)

    def __init__(self):
        super(DataModel, self).__init__()
        self.id_counter = 0
        self._volumes = {}
        self._data = {}
        self._vectors = {}

    def change_vol_name(self, old_name, new_name):
        # Only work on image volumes for now
        if self._volumes.get(old_name):
            # Change the dictionary key entry
            self._volumes[new_name] = self._volumes.pop(old_name)
            # Change the id on the object
            self._volumes[new_name].name = new_name

    def set_interpolation(self, onoff):
        for vol in self._volumes.values():
            vol.set_interpolation(onoff)

    def clear_data(self):
        keys = list(self._volumes.keys())
        for k in keys:
            del self._volumes[k]
        self._volumes = {}
        self._data = {}

    def volume_id_list(self, sort=True):
        if sort: # Not sure if we need this
            return sorted([id_ for id_ in self._volumes])
        else:
            return [id_ for id_ in self._volumes]

    def data_id_list(self):
        return sorted([id_ for id_ in self._data])

    def vector_id_list(self):
        return sorted([id_ for id_ in self._vectors])

        for key in self._data.keys():
            self._data[key].destroy()

    def all_volumes(self):
        return [vol for vol in self._volumes.values()]

    def getvol(self, id_):
        # bodge. should merge vols and data, as they have unique ids
        vol = None
        if id_ == 'None':
            return 'None'
        try:
            vol = self._volumes[id_]
        except KeyError:
            pass

        if not vol:
            try:
                vol = self._data[id_]
            except KeyError:
                pass
        if not vol:
            try:
                vol = self._vectors[id_]
            except KeyError:
                return "None"  # Need to do something else here, like logging
        return vol

    def getdata(self, id_):
        if id_ == 'None':
            return 'None'
        return self._data[id_]

    def load_image_series(self, series_paths, memory_map):
        volpath = str(series_paths[0])
        n = os.path.basename(volpath)
        unique_name = self.create_unique_name(n)
        vol = ImageSeriesVolume(series_paths, self, 'series', memory_map)
        vol.name = unique_name
        self._volumes[vol.name] = vol
        self.id_counter += 1

    def load_annotation(self, ann_path):
        """
        Load annotations from an IMPC xml file.

        Parameters
        ----------
        ann_path: str
            path to xml annotation file

        Returns
        -------

        """
        # Load in data from xml
        try:
            centerID, pipeline, project, doe, ex_id, spec_id, proc_id, \
            simple_and_series_params, procedure_metadata = load_xml(ann_path)
        except IOError as e:
            print("Cannot read xml file {}\n".format(ann_path, e))
            error_dialog(None, 'File read error', "Problem reading annotaitons file\n{}".format(ann_path))
            return

        # try to find a corresponding procedure_metadata.yaml file
        ann_dir = os.path.split(ann_path)[0]
        procedure_metadata_file = os.path.join(ann_dir, PROCEDURE_METADATA)

        if not os.path.isfile(procedure_metadata_file):
            vol = None  # Should also check if annotation options have been loaded
        else:
            vol_id = os.path.basename(ann_dir)  # The annotation directory is the same name as the annotated volume
            vol = self._volumes.get(vol_id)

        if not vol:
            return "Could not load annotation: {}. Not able to find loaded volume with same id".format(vol_id)

        vol.annotations.clear()
        # Get the dict that contains the available options for a given center/stage
        annotation_date_param_id = get_annotator_id_and_date(proc_id)[1]
        ann_date = [x[1] for x in procedure_metadata if x[0] == annotation_date_param_id]
        ann_date = ann_date[0]
        vol.annotations.annotation_date = ann_date

        default_opts = centre_stage_options.opts
        stage = get_stage_and_modality(proc_id, centerID)
        ######################################
        # This all needs moving into Annotations
        # Set the xml file path which is where it will get resaved to
        vol.annotations.saved_xml_fname = ann_path

        # Get all the simpleParameter entries form the xml file
        for xml_param, xml_data in simple_and_series_params.items():
            option = xml_data['option']
            xyz = xml_data.get('xyz')
            if xyz:
                x, y, z = [int(i) for i in xyz]
            else:
                x = y = z = None
            dims = vol.shape_xyz()

            # Some of the data needed to create an annotation object is not recorded in the XML output
            # So we need to load that from the center annotation options file
            for center, default_data in default_opts['centers'].items():
                if default_data['short_name'] == centerID:
                    params = default_data['procedures'][proc_id]['parameters']

                    for param_id, default_param_info in params.items():
                        if param_id == xml_param:
                            name = default_param_info['name']
                            options = default_opts['available_options'][
                                default_param_info['options']]  # available options for this parameter
                            order = default_param_info['order']
                            is_mandatory = default_param_info['mandatory']

                            vol.annotations.add_impc_annotation(x, y, z, xml_param, name, options, option,
                                                                stage,
                                                                order, is_mandatory, dims)
        vol.annotations._load_done_status()



    # def load_annotation(self, ann_path):
    #     """
    #     Load annotations from an IMPC xml file.
    #
    #     Parameters
    #     ----------
    #     ann_path: str
    #         path to xml annotation file
    #
    #     Returns
    #     -------
    #
    #     """
    #     # Load in data from xml
    #     centerID, pipeline, project, doe, ex_id, spec_id, proc_id, \
    #     simple_and_series_params, procedure_metadata = load_xml(ann_path)
    #
    #     # try to find a corresponding procedure_metadata.yaml file
    #     ann_dir = os.path.split(ann_path)[0]
    #     procedure_metadata_file = os.path.join(ann_dir, PROCEDURE_METADATA)
    #     if not os.path.isfile(procedure_metadata_file):
    #         vol = None  # Should also check if annotation options have been loaded
    #     else:
    #         vol_id = os.path.basename(ann_dir) # The annotation directory is the same name as the annotated volume
    #         vol = self._volumes.get(vol_id)
    #
    #     if vol:
    #         # Get the dict that contains the available options for a given center/stage
    #         default_opts = centre_stage_options.opts
    #         stage = get_stage_from_proc_id(proc_id, centerID)
    #
    #         # Get all the simpleParameter entries form the xml file
    #         for xml_param, xml_data in simple_and_series_params.items():
    #             option = xml_data['option']
    #             xyz = xml_data.get('xyz')
    #             if xyz:
    #                 x, y, z = [int(i) for i in xyz]
    #             else:
    #                 x = y = z = None
    #             dims = vol.shape_xyz()
    #
    #             # Some of the data neded to crate an annotation object is not recorded in the XML output
    #             # So we need to load that from the center annotation options file
    #             for center, default_data in default_opts['centers'].items():
    #                 if default_data['short_name'] == centerID:
    #                     params = default_data['stages'][stage]['parameters']
    #
    #                     for param_id, default_param_info in params.items():
    #                         if param_id == xml_param:
    #                             name = default_param_info['name']
    #                             options = default_opts['available_options'][default_param_info['options']]# available options for this parameter
    #                             order = default_param_info['options']
    #                             is_mandatory = default_param_info['mandatory']
    #
    #                             vol.annotations.add_impc_annotation(x, y, z, xml_param, name, options, option, stage,
    #                                                         order, is_mandatory, dims)
    #
    #     else:
    #         return "Could not load annotation: {}. Not able to find loaded volume with same id".format(vol_id)
    #     return None

    def add_volume(self, volpath, data_type, memory_map, fdr_thresholds=False):
        """
        Load a volume into a subclass of a Volume object
        Parameters
        ----------
        volpath: str
        data_type: str
        memory_map: bool
        fdr_thresholds: fdict
            q -> t statistic mappings
                {0.01: 3.4,
                0.05:, 3.1}
        """

        if data_type != 'virtual_stack':
            volpath = str(volpath)
            n = os.path.basename(volpath)
            unique_name = self.create_unique_name(n)
        else:
            n = os.path.basename(os.path.split(volpath[0])[0])
            unique_name = self.create_unique_name(n)

        if data_type == 'heatmap':
            vol = HeatmapVolume(volpath, self, 'heatmap')
            if fdr_thresholds or fdr_thresholds is None:
                vol.fdr_thresholds = fdr_thresholds
            vol.name = unique_name
            self._data[vol.name] = vol
        elif data_type == 'vol':
            vol = ImageVolume(volpath, self, 'volume', memory_map)
            vol.name = unique_name
            self._volumes[vol.name] = vol
        elif data_type == 'virtual_stack':
            vol = VirtualStackVolume(volpath, self, 'virtual_stack', memory_map)
            vol.name = unique_name
            self._volumes[vol.name] = vol
        elif data_type == 'vector':
            vol = VectorVolume(volpath, self, 'vector')
            vol.name = unique_name
            self._vectors[vol.name] = vol

        self.id_counter += 1
        self.data_changed_signal.emit()

    def create_unique_name(self, name):
        """
        Create a unique name for each volume. If it already exists, append a digit in a bracket to it
        :param name:
        :return:
        """
        name = os.path.splitext(name)[0]
        if name not in self._volumes and name not in self._data and name not in self._vectors:
            return name
        else:
            for i in range(1, 100):
                new_name = '{}({})'.format(name, i)
                if new_name not in self._volumes and new_name not in self._data:
                    return new_name

    def write_temporary_annotations_metadata(self):
        """

        Returns
        -------

        """
        from os.path import join
        for id_, vol in self._volumes.items():

            if vol.annotations.annotation_dir:

                # Check for previous done list
                done_file = join(vol.annotations.annotation_dir, ANNOTATION_DONE_METADATA_FILE)
                done_status = {}

                for ann in vol.annotations:
                    done_status[ann.term] = ann.looked_at

                with open(done_file, 'w') as fh:
                    fh.write(yaml.dump(done_status))



from vpv.common import Orientation
import numpy as np
from vpv.display.slice_view_widget import SliceWidget

flip_to_axial_order = {
    # The order of the source dimensions after mapping to axial space
    # For example [2, 1, 0] reverses the dimension ordering as we are using

    Orientation.axial:    [0, 1, 2],  # keep same
    Orientation.coronal:  [0, 2, 1],
    Orientation.sagittal: [1, 2, 0]
}

flip_from_axial_order = {
    Orientation.axial:    [0, 1, 2],  # Keep same
    Orientation.coronal:  [0, 2, 1],
    Orientation.sagittal: [2, 0, 1]
}


class Coordinate_mapper(object):
    """Map coordinates between views and volumes
    """
    def __init__(self, views: dict, saved_flip_info: dict):
        self.views = views  # Get a reference to the views dict that contains the slice view widgets
        self.flip_info = saved_flip_info

    def view_to_volume(self, x: int, y: int, z: int, src_ori: Orientation, dims: list, from_saved=False) -> tuple:
        """
        Given coordinates from a slice view, convert to actual coordinates in the correct volume space
        This is required as we do some inversion of the order of slices as they come off the volumes to show
        a view that the IMPC annotators like.

        Currently not working if there are no Slice views that are in axial view

        Parameters
        ----------
        x: int
        y: int
        z: int
        src_ori: Orientation
            The orientation of the source view
        dims: list
            dimensions of source volume xyz
        from_saved: bool


        Returns
        -------
        (x, y, z)

        Notes
        -----

        This maps the view coordinates from any view to the normal axial view. Then adds correction for the inverted
        slice ordering

        Problem: If the impc_vie is active the view_to_view() inverts x when mapping between
        axial and the other views so that cross hairs work. this breaks the mouse
        position labels

        """

        if not from_saved:  # If from saved list we don't need to do any flipping on the points
            # Flip the dimensions so thay match the view

            dims_order = flip_from_axial_order[src_ori]
            new_dims = [j for _, j in sorted(zip(dims_order, dims), key=lambda pair: pair[0])]


            if self.flip_info[src_ori.name]['y']:
                y = new_dims[1] - y

            if not self.flip_info[src_ori.name]['x']:  # 280319. x is flipped by default in every view for RAS
                x = new_dims[0] - x

            if self.flip_info[src_ori.name]['z']:
                z = new_dims[2] - z

        order = flip_to_axial_order[src_ori]
        xa, ya, za = [j for _, j in sorted(zip(order, [x, y, z]), key=lambda pair: pair[0])]

        return xa, ya, za

    def view_to_view(self, x, y, z, src_ori, dest_ori, dims, from_saved=False):
        """
        Given a coordinate on one view with a given orientation as well as horizontal flip and slice ordering status,
        map to another view with a given orientation , flip and ordering.
        Parameters
        ----------
        x: int
            the x position in the view
        y: int
            the y position in the view
        z: int
            the slice index
        src_view: [SliceWidget, None]
            The slice view widget that the corrdinate is mapped from. If none assum nornal volume space (z == axial)
        dest_view: SliceWidget
            The slice view widget to map the coordinates to
        rev: bool
            whether to return the inverted point (len(dimension) - point)
        Returns
        -------
        tuple
            (x,y,idx)

        """
        # Get the points in volume space
        axial_space_points = self.view_to_volume(x, y, z, src_ori, dims, from_saved)

        # And map to the destination space

        order = flip_from_axial_order[dest_ori]
        dest_points = [j for _, j in sorted(zip(order, axial_space_points), key=lambda pair: pair[0])]

        dims_order = flip_from_axial_order[dest_ori]
        new_dims = [j for _, j in sorted(zip(dims_order, dims), key=lambda pair: pair[0])]

        if self.flip_info[dest_ori.name]['y']:
            dest_points[1] = new_dims[1] - dest_points[1]

        if not self.flip_info[dest_ori.name]['x']:
            dest_points[0] = new_dims[0] - dest_points[0]

        if self.flip_info[dest_ori.name]['z']:
            dest_points[2] = new_dims[2] - dest_points[2]

        return dest_points

    def roi_to_view(self, xx, yy, zz):

        """
        Given an roi is volume coordinates (z == axial), map the roi to other orthogonal views
        Parameters
        ----------
        xx: tuple
        yy: tuple
        zz: tuple

        Returns
        -------
        tuple
            Mapped coordinates ((x,x), (y,y), (z,z)
        """
        z = [int(i) for i in zz]
        y = [int(i) for i in yy]
        x = [int(i) for i in xx]

        for dest_view in self.views.values():
            # First map the annotation marker between views
            dest_dims = dest_view.main_volume.shape_xyz()

            points_1 = self.view_to_view(x[0], y[0], z[0], Orientation.axial, dest_view.orientation,
                                         dest_dims, from_saved=True)

            points_2 = self.view_to_view(x[1], y[1], z[1], Orientation.axial, dest_view.orientation,
                                         dest_dims, from_saved=True)

            midslice = int(np.mean([points_1[2], points_2[2]]))

            w = abs(points_1[0] - points_2[0])
            h = abs(points_1[1] - points_2[1])
            dest_view.set_slice(midslice)

            # we need to set the start of the roi as the lowest valued x and y. These change if flips occur
            new_x = min((points_1[0], points_2[0]))
            new_y = min((points_1[1], points_2[1]))
            dest_view.set_roi(new_x, new_y, w, h)


# def convert_volume(vol, direction):
#     return vol
#     if direction == (1, 0, 0, 0, 1, 0, 0, 0, 1): # LPS
#         vol = vol[:, ::-1, :]
#     if direction == (-1, 0, 0, 0, -1, 0, 0, 0, 1): # RAS
#         print("ras?")
#         vol = vol
#         # vol = vol[:, :, ::-1]
#
#     return vol



from PyQt5 import QtCore, Qt
from vpv.utils.lookup_tables import Lut
import numpy as np


class Layer(Qt.QObject):

    volume_label_signal = QtCore.pyqtSignal(str)

    def __init__(self, parent, layer_type, model):
        """
        The base class that controls the display of of one layer of data. Is extended to display heatmap data and vector
        data. The main function is to get data into a pytgraph ImageItem object which is then displayed.

        Attributes
        ----------
        model: model.model
            contains all the data
        parent: display.SliceWidget
            the parent view widget
        image_item: pyqtgrapgh.ImageItem
            The image to be displayed is put in here

        """
        super(Layer, self).__init__()
        # If it's a data volume, dont add slider, add number box to filter q-values
        self.model = model
        self.parent = parent
        self.image_items = []
        self.image_item = None  # Set in the subclasses
        self.vol = None  # When setting to "None" in the view manager we get problems. Needs rewriting
        self.layer_type = layer_type # this should be unique
        self.lt = Lut()
        self.isvisible = True
        self.opacity = 1.0
        self._show_labels = [0]  # show all labels to start with (if label layer)

    @property
    def show_labels(self):
        return self._show_labels

    @show_labels.setter
    def show_labels(self, labels):
        self._show_labels = labels

    def clear(self):
        """
        To clear the pg.imageitem set the slice to all zeros
        Returns
        -------

        """
        slice_ = np.copy(self.image_item.image)
        try:
            slice_[:] = 0
        except IndexError:
            return
        self.image_item.setImage(slice_)

    def update(self, auto_levels=False):
        """
        Sets the orientation and centres the image at the mid-slice
        :param orientation:
        :param auto_levels:
        :return:
        """
        if self.vol and self.vol != 'None':
            self.image_item.setLevels(self.vol.levels, update=False)
            self.image_item.setLookupTable(self.lut[0])
            self.reload()

    def reload(self):
        """
        reload the current image. For example when the orientation has changed
        """
        self.set_slice(self.parent.current_slice_idx)

    def set_volume(self, volname, initial=False):
        """
        :param vol, Volume object from model.py
        """

        if volname == "None" or volname is None:
            self.volume_label_signal.emit("None")
            self.clear()  # This clears the image
            self.vol = None
            return

        self.volume_label_signal.emit(volname)

        self.vol = self.model.getvol(volname)
        if not self.vol:
            print(f'cannot find vol: {volname}')
            return

        # self.set_series_slider()

        orientation = self.parent.orientation
        self.name = self.vol.name
        self.image_item.setZValue(self.layer_type.value)

        dim_len = self.vol.dimension_length(orientation)

        # Try setting the new volume to the slice index of the current volume. If out of bounds, set to midslice
        if self.parent.current_slice_idx + 1 > dim_len or initial:
            slice_indx = dim_len / 2
        else:
            slice_indx = self.parent.current_slice_idx
        # slice_ = self.vol.get_data(orientation, int(slice_indx))  # DELETE
        self.parent.set_slice_slider(dim_len, slice_indx)

        # This fixes problem with linked zooming
        if initial:
            self.parent.viewbox.autoRange()
        self.parent.scalebar.updateBar()

    def set_slice(self, index: int):
        opacity = self.opacity if self.isvisible else 0.0
        flip_x, flip_y, flip_z = self.parent.get_flips()

        if self.vol:

            try:
                slice_ = self.vol.get_data(self.parent.orientation, index,
                                                       flip_x, flip_z, flip_y)

                if self._show_labels != [0]:
                    slice_ = np.copy(slice_)
                    slice_[~np.isin(slice_, self._show_labels)] = 0

                self.image_item.setImage(slice_, autoLevels=False, opacity=opacity)

            except IndexError as e:
                print(e)

    def set_series_slider(self):
        if not self.vol or self. vol == 'None':
            return
        if self.vol.data_type == 'series':
            num_vols_in_series = len(self.vol.images)
            self.parent.ui.seriesSlider.setRange(0, num_vols_in_series - 1)
            self.parent.ui.seriesSlider.valueChanged.connect(self.series_slider_changed)
            self.parent.ui.seriesSlider.show()
            self.parent.ui.labelImageSeriesNumber.show()
        else:
            self.parent.ui.seriesSlider.hide()
            self.parent.ui.labelImageSeriesNumber.hide()

    def series_slider_changed(self, idx):
        self.vol.set_image(idx)
        self.parent.ui.labelImageSeriesNumber.setText(str(idx))
        self.reload()

    def set_visibility(self, isvisible: bool):
        self.isvisible = isvisible

    def set_opacity(self, opacity: float):
        self.opacity = opacity



"""
18/12/18
Convert the XML annotations to csv
"""

from pathlib import Path
import pandas as pd
from vpv.annotations.impc_xml import load_xml
import yaml

from collections import defaultdict


xml_dir = Path('/home/neil/Desktop/xml_to_csv')
outfile = '/home/neil/Desktop/181218_xml_annotations_to_csv.csv'


term_mapping_file = '/vpv_viewer/annotations/options/e15_5/ucd_terms.yaml'
with open(term_mapping_file) as fh:
    terms = yaml.load(fh)

map_ = {}
for _, p in terms['parameters'].items():
    param_id = p['impc_id']
    name = p['name']
    map_[param_id] = name


all_spec = []

for xml_file in xml_dir.iterdir():
    r = load_xml(str(xml_file))
    print(r)

    spec_records = {}

    spec_id = r[5]

    for param_id, options in r[7].items():
        option = options['option']
        param_name = map_[param_id]
        spec_records[param_name] = option

    df = pd.DataFrame.from_dict(spec_records, orient='index', columns=['option'])
    df.rename(columns={'option': spec_id}, inplace=True)

    all_spec.append(df)

o = pd.concat(all_spec, axis=1)

o.to_csv(outfile)


from nose.tools import assert_equals, nottest
import sys
from os.path import join, dirname
sys.path.insert(0, join(dirname(__file__), '..'))
from data_manager import ColorScaleBar
from PyQt5 import QtGui, QtWidgets
from lookup_tables import Lut


def test_color_bar():
    """
    Just runs the color scale bar code. Not a proper test as such. You have to tell whether it's passed by looking
    at the image produced

    Used in development as it's quicker than running an instance of vpv_viewer each time
    """
    lut = Lut()
    hotred = lut._hot_red_blue()[0]
    hotblue = lut._hot_red_blue()[1]

    app = QtGui.QApplication(sys.argv)
    w = QtGui.QWidget()
    w.resize(250, 150)
    w.move(300, 300)
    w.setWindowTitle('color bar test')
    layout =  QtWidgets.QVBoxLayout()
    w.setLayout(layout)
    cb = ColorScaleBar(layout, hotblue, hotred)
    cb.show()
    cb.update(40, 10, -37, -3)

    w.show()
    sys.exit(app.exec_())
